from fink_utils.slack_bot.rich_text.rich_text import RichElement
from fink_utils.slack_bot.rich_text.rich_text_element import RichTextElement
from typing import List, Union
from typing_extensions import Self


class SectionElement(RichElement):
    def __init__(self) -> None:
        """
        A class representing a section of a rich text
        """
        super().__init__()
        self.section = {"type": "rich_text_section", "elements": []}

    def add_elements(
        self, elements: Union[RichTextElement, List[RichTextElement]]
    ) -> Self:
        if type(elements) is list:
            self.section["elements"] += [el.get_element() for el in elements]
        else:
            self.section["elements"].append(elements.get_element())
        return self

    def get_element(self) -> dict:
        return self.section



from abc import ABC, abstractmethod
from enum import Enum


class RichTextStyle(Enum):
    BOLD = "bold"
    ITALIC = "italic"
    STRIKE = "strike"
    CODE = "code"


class RichTextElement(ABC):
    @abstractmethod
    def __init__(self) -> None:
        pass

    @abstractmethod
    def get_element(self):
        pass


class Emoji(RichTextElement):
    def __init__(self, emoji_name) -> None:
        super().__init__()
        self.emoji = {"type": "emoji", "name": emoji_name}

    def get_element(self):
        return self.emoji


class Link(RichTextElement):
    def __init__(
        self,
        url: str,
        text: str = None,
        unsafe: bool = None,
        style: RichTextStyle = None,
    ) -> None:
        """
        A link element

        Parameters
        ----------
        url : str
            url
        text : str, optional
            text to show instead of the url, if not provided, show the url, by default None
        unsafe : bool, optional
            indicates whether the link is safe, by default None
        style : RichTextStyle, optional
            link style, by default None
        """
        super().__init__()
        self.link = {"type": "link", "url": url}
        if text is not None:
            self.link["text"] = text
        if unsafe is not None:
            self.link["unsafe"] = unsafe
        if style is not None:
            self.link["style"] = style.value

    def get_element(self):
        return self.link


class Text(RichTextElement):
    def __init__(self, text: str, style: RichTextStyle = None) -> None:
        """
        A text element

        Parameters
        ----------
        text : str
            text
        style : RichTextStyle, optional
            text style, by default None
        """
        super().__init__()
        self.text = {"type": "text", "text": text}
        if style is not None:
            self.link["style"] = style.value

    def get_element(self):
        return self.text



import os
import time
import io

from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError
from typing import List

from fink_utils.logging.logs import init_logging, LoggerNewLine
from fink_utils.slack_bot.msg_builder import Message

import json


def init_slackbot(logger: LoggerNewLine = None) -> WebClient:
    """
    Initialize a slack bot

    Returns
    -------
    WebClient
        the web client used by the bot
    """
    try:
        token_slack = os.environ["FINK_SLACK_TOKEN"]
    except KeyError as e:
        if logger is None:
            logger = init_logging()
        logger.error("FINK_SLACK_TOKEN environement variable not found !!", exc_info=1)
        raise e
    client = WebClient(token=token_slack)
    return client


def post_files_on_slack(
    slack_client: WebClient,
    file_list: List[io.BytesIO],
    title_list: List[str],
    sleep_delay: int = 1,
) -> List[str]:
    """
    Post the files in bytes format on slack with the associated title.

    Parameters
    ----------
    slack_client : WebClient
        slack bot client
    file_list : list[io.BytesIO]
        files to upload on slack in bytes format
    title_list : list[str]
        title to give to the files, should be in the same order than the file_list
    sleep_delay : int, optional
        delay to wait before executing the instruction after the upload, by default 1

    Returns
    -------
    list[str]
        list of string containing the link to refers to the files upload on slack in a msg
    """
    results = slack_client.files_upload_v2(
        file_uploads=[
            {
                "file": file,
                "title": title,
            }
            for file, title in zip(file_list, title_list)
        ]
    )
    time.sleep(sleep_delay)
    return [
        f"https://files.slack.com/files-tmb/{el['user_team']}-{el['id']}/"
        for el in results["files"]
    ]


def post_msg_on_slack(
    webclient: WebClient,
    channel: str,
    msg: List[Message],
    sleep_delay: int = 1,
    logger: LoggerNewLine = None,
    verbose: bool = False,
):
    """
    Send a msg on the specified slack channel

    Parameters
    ----------
    webclient : WebClient
        slack bot client
    channel : str
        the channel where will be posted the message
    msg : list
        list of message to post on slack, each string in the list will be a single post.
    sleep_delay : int, optional
        delay to wait between the message, by default 1
    logger : _type_, optional
        logger used to print logs, by default None
    verbose : bool, optional
        if true, print logs between the message, by default False

    * Notes:

    Before sending message on slack, check that the Fink bot have been added to the targeted channel.

    Examples
    --------
    see bot_test.py
    """
    if verbose and logger is None:
        logger = init_logging()
    try:
        for tmp_msg in msg:
            json_p = json.dumps(tmp_msg.blocks["blocks"])
            webclient.chat_postMessage(
                channel=channel, text="error with msg blocks", blocks=json_p
            )
            if verbose:
                logger.debug("Post msg on slack successfull")
            time.sleep(sleep_delay)
    except SlackApiError as e:
        if e.response["ok"] is False:
            logger.error("Post slack msg error", exc_info=1)



# Copyright 2020-2024 AstroLab Software
# Author: Julien Peloton
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pyspark.sql.functions import pandas_udf, PandasUDFType
from pyspark.sql.types import TimestampType
from pyspark.sql import SparkSession

import numpy as np
import pandas as pd
from astropy.time import Time

# from fink_utils.test.tester import spark_unit_tests_broker


@pandas_udf(TimestampType(), PandasUDFType.SCALAR)
def convert_to_millitime(jd: pd.Series, format=None, now=None):
    """Convert date into unix milliseconds (long)

    Parameters
    ----------
    jd: double
        Julian date
    format: str, optional
        Astropy time format, e.g. jd, mjd, ... Default is jd.
    now: boolean, optional
        If True, return the current time. Default is False

    Returns
    -------
    out: pd.Series
        Unix milliseconds in UTC

    Examples
    --------
    >>> from fink_broker.sparkUtils import load_parquet_files
    >>> df = load_parquet_files("online/raw")
    >>> df = df.withColumn('millis', convert_to_millitime(df['candidate.jd']))
    >>> pdf = df.select('millis').toPandas()
    >>> import pyspark.sql.functions as F
    >>> df = df.withColumn('millis', convert_to_millitime(
    ...     df['candidate.jd'], F.lit('jd'), F.lit(True)))
    >>> pdf = df.select('millis').toPandas()
    """
    if format is None:
        formatval = "jd"
    else:
        formatval = format.to_numpy()[0]

    if now is not None:
        times = [Time.now().to_datetime()] * len(jd)
    else:
        times = Time(jd.to_numpy(), format=formatval).to_datetime()

    return pd.Series(times)


@pandas_udf(TimestampType(), PandasUDFType.SCALAR)
def convert_to_datetime(jd: pd.Series, format=None) -> pd.Series:
    """Convert date into datetime (timestamp)

    Be careful if you are using this outside Fink. First, you need to check
    you timezone defined in Spark:
    ```
    spark.conf.get("spark.sql.session.timeZone")
    ```
    If this is something else than UTC, then change it:
    ```
    spark.conf.set("spark.sql.session.timeZone", 'UTC')
    ```

    Parameters
    ----------
    jd: double
        Julian date
    format: str
        Astropy time format, e.g. jd, mjd, ...

    Returns
    -------
    out: datetime
        Datetime object in UTC

    Examples
    --------
    >>> from fink_broker.sparkUtils import load_parquet_files
    >>> df = load_parquet_files("online/raw")
    >>> df = df.withColumn('datetime', convert_to_datetime(df['candidate.jd']))
    >>> pdf = df.select('datetime').toPandas()
    """
    if format is None:
        formatval = "jd"
    else:
        formatval = format.to_numpy()[0]

    return pd.Series(Time(jd.to_numpy(), format=formatval).to_datetime())


def numPart(df, partition_size=128.0):  # noqa: N802
    """Compute the idle number of partitions of a DataFrame based on its size.

    Parameters
    ----------
    partition_size: float
        Size of a partition in MB

    Returns
    -------
    numpart: int
        Number of partitions of size `partition_size` based
        on the DataFrame size
    partition_size: float, optional
        Size of output partitions in MB

    Examples
    --------
    >>> from fink_broker.sparkUtils import load_parquet_files
    >>> df = load_parquet_files("online/raw")
    >>> numpart = numPart(df, partition_size=128.)
    >>> print(numpart)
    1
    """
    # Grab the running Spark Session,
    # otherwise create it.
    spark = SparkSession.builder.getOrCreate()

    b = (
        spark._jsparkSession.sessionState()
        .executePlan(df._jdf.queryExecution().logical())
        .optimizedPlan()
        .stats()
        .sizeInBytes()
    )

    # Convert in MB
    b_mb = b / 1024.0 / 1024.0

    # Ceil it
    numpart = int(np.ceil(b_mb / partition_size))

    return numpart


# if __name__ == "__main__":
#     """ Execute the test suite with SparkSession initialised """

#     # Run the Spark test suite
#     spark_unit_tests_broker(globals())



from fink_utils.slack_bot.section import TypeText


class Image:
    def __init__(self, image_url: str, alt_text: str) -> None:
        """
        Instanciate an image to upload on slack

        Parameters
        ----------
        image_url : str
            url of the image
        alt_text : str
            text of the image
        """
        self.image = {"type": "image", "image_url": image_url, "alt_text": alt_text}

    def add_title(self, type_text: TypeText, title: str, allow_emoji: bool = False):
        """
        Add a title to the image

        Parameters
        ----------
        type_text : TypeText
            type of the title
        title : str
            text of the title
        allow_emoji : bool, optional
            if True allow the emoji in the text, by default False
        """
        self.image["title"] = {
            "type": type_text.value,
            "text": title,
            "emoji": allow_emoji,
        }

    def get_element(self):
        return self.image



# Copyright 2020-2024 AstroLab Software
# Author: Julien Peloton
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pyspark.sql.functions import pandas_udf, PandasUDFType
from pyspark.sql.types import StringType, LongType

import pandas as pd
import numpy as np
import healpy as hp


def dec2theta(dec: float) -> float:
    """Convert Dec (deg) to theta (rad)"""
    return np.pi / 2.0 - np.pi / 180.0 * dec


def ra2phi(ra: float) -> float:
    """Convert RA (deg) to phi (rad)"""
    return np.pi / 180.0 * ra


@pandas_udf(LongType(), PandasUDFType.SCALAR)
def ang2pix(ra: pd.Series, dec: pd.Series, nside: pd.Series) -> pd.Series:
    """Compute pixel number at given nside

    Parameters
    ----------
    ra: float
        Spark column containing RA (float)
    dec: float
        Spark column containing RA (float)
    nside: int
        Spark column containing nside

    Returns
    -------
    out: long
        Spark column containing pixel number

    Examples
    --------
    >>> from fink_broker.sparkUtils import load_parquet_files
    >>> df = load_parquet_files(ztf_alert_sample)
    >>> df_index = df.withColumn(
    ...     'p',
    ...     ang2pix(df['candidate.ra'], df['candidate.dec'], F.lit(256))
    ... )
    >>> df_index.select('p').take(1)[0][0] > 0
    True
    """
    return pd.Series(
        hp.ang2pix(
            nside.to_numpy()[0], dec2theta(dec.to_numpy()), ra2phi(ra.to_numpy())
        )
    )


@pandas_udf(StringType(), PandasUDFType.SCALAR)
def ang2pix_array(ra: pd.Series, dec: pd.Series, nside: pd.Series) -> pd.Series:
    """Return a col string with the pixel numbers corresponding to the nsides pix@nside[0]_pix@nside[1]_...etc

    Parameters
    ----------
    ra: float
        Spark column containing RA (float)
    dec: float
        Spark column containing RA (float)
    nside: list
        Spark column containing list of nside

    Returns
    -------
    out: str
        Spark column containing _ separated pixel values

    Examples
    --------
    >>> from fink_broker.sparkUtils import load_parquet_files
    >>> df = load_parquet_files(ztf_alert_sample)
    >>> nsides = F.array([F.lit(256), F.lit(4096), F.lit(131072)])
    >>> df_index = df.withColumn(
    ...     'p',
    ...     ang2pix_array(df['candidate.ra'], df['candidate.dec'], nsides)
    ... )
    >>> l = len(df_index.select('p').take(1)[0][0].split('_'))
    >>> print(l)
    3
    """
    pixs = [
        hp.ang2pix(int(nside_), dec2theta(dec.to_numpy()), ra2phi(ra.to_numpy()))
        for nside_ in nside.to_numpy()[0]
    ]

    to_return = ["_".join(list(np.array(i, dtype=str))) for i in np.transpose(pixs)]

    return pd.Series(to_return)


# if __name__ == "__main__":
#     """ Execute the test suite with SparkSession initialised """

#     globs = globals()
#     root = os.environ['FINK_HOME']
#     globs["ztf_alert_sample"] = os.path.join(
#         root, "online/raw")

#     globs['elasticc_alert_sample'] = os.path.join(
#         root, "elasticc_parquet")

#     # Run the Spark test suite
#     spark_unit_tests(globs)



# Copyright 2024 AstroLab Software
# Author: Julien Peloton
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utils for fitting asteroid synodic period of rotation"""

from fink_utils.sso.spins import (
    estimate_sso_params,
    func_sshg1g2,
    func_hg1g2_with_spin,
    func_hg1g2,
    func_hg12,
    func_hg,
)
from fink_utils.sso.utils import compute_light_travel_correction
from astropy.timeseries import LombScargleMultiband

import requests
import numpy as np
import io
import pandas as pd
from line_profiler import profile
import nifty_ls  # noqa: F401

import logging

from fink_utils.test.tester import regular_unit_tests

_LOG = logging.getLogger(__name__)


def extract_physical_parameters(pdf, flavor):
    """Fit for phase curve parameters

    Parameters
    ----------
    pdf: pandas DataFrame
        DataFrame with SSO data from Fink REST API
    flavor: str
        Model flavor: SHG1G2, HG1G2, HG12, or HG

    Returns
    -------
    outdic: dict
        Dictionary containing reduced chi2, and estimated parameters and
        error on each parameters.
    """
    if flavor == "HG1G2":
        bounds = (
            [0, 0, 0],
            [30, 1, 1],
        )
        p0 = [15.0, 0.15, 0.15]
    elif flavor == "HG12":
        bounds = (
            [0, 0],
            [30, 1],
        )
        p0 = [15.0, 0.15]
    elif flavor == "HG":
        bounds = (
            [0, 0],
            [30, 1],
        )
        p0 = [15.0, 0.15]
    elif flavor == "SHG1G2":
        bounds = (
            [0, 0, 0, 3e-1, 0, -np.pi / 2],
            [30, 1, 1, 1, 2 * np.pi, np.pi / 2],
        )
        p0 = [15.0, 0.15, 0.15, 0.8, np.pi, 0.0]

    outdic = estimate_sso_params(
        magpsf_red=pdf["i:magpsf_red"].to_numpy(),
        sigmapsf=pdf["i:sigmapsf"].to_numpy(),
        phase=np.deg2rad(pdf["Phase"].to_numpy()),
        filters=pdf["i:fid"].to_numpy(),
        ra=np.deg2rad(pdf["i:ra"].to_numpy()),
        dec=np.deg2rad(pdf["i:dec"].to_numpy()),
        p0=p0,
        bounds=bounds,
        model=flavor,
        normalise_to_V=False,
    )

    return outdic


def compute_residuals(pdf, flavor, phyparam):
    """Compute residuals between data and predictions

    Parameters
    ----------
    pdf: pandas DataFrame
        DataFrame with SSO data from Fink REST API
    flavor: str
        Model flavor: SHG1G2, HG1G2, HG12, or HG
    phyparam: dict
        Dictionary containing reduced chi2, and estimated parameters and
        error on each parameters.

    Returns
    -------
    pd.Series
        Series containing `observation - model` in magnitude
    """
    pdf["preds"] = 0.0
    for filtnum in pdf["i:fid"].unique():
        if filtnum == 3:
            continue
        cond = pdf["i:fid"] == filtnum

        if flavor == "SSHG1G2":
            pha = [
                np.deg2rad(pdf["Phase"][cond]),
                np.deg2rad(pdf["i:ra"][cond]),
                np.deg2rad(pdf["i:dec"][cond]),
            ]
            preds = func_sshg1g2(
                pha,
                phyparam["H_{}".format(filtnum)],
                phyparam["G1_{}".format(filtnum)],
                phyparam["G2_{}".format(filtnum)],
                np.deg2rad(phyparam["alpha0"]),
                np.deg2rad(phyparam["delta0"]),
                phyparam["period"],
                phyparam["a_b"],
                phyparam["a_c"],
                phyparam["phi0"],
            )
        elif flavor == "SHG1G2":
            pha = [
                np.deg2rad(pdf["Phase"][cond]),
                np.deg2rad(pdf["i:ra"][cond]),
                np.deg2rad(pdf["i:dec"][cond]),
            ]
            preds = func_hg1g2_with_spin(
                pha,
                phyparam["H_{}".format(filtnum)],
                phyparam["G1_{}".format(filtnum)],
                phyparam["G2_{}".format(filtnum)],
                phyparam["R"],
                np.deg2rad(phyparam["alpha0"]),
                np.deg2rad(phyparam["delta0"]),
            )
        elif flavor == "HG":
            preds = func_hg(
                np.deg2rad(pdf["Phase"][cond]),
                phyparam["H_{}".format(filtnum)],
                phyparam["G_{}".format(filtnum)],
            )
        elif flavor == "HG12":
            preds = func_hg12(
                np.deg2rad(pdf["Phase"][cond]),
                phyparam["H_{}".format(filtnum)],
                phyparam["G12_{}".format(filtnum)],
            )
        elif flavor == "HG1G2":
            preds = func_hg1g2(
                np.deg2rad(pdf["Phase"][cond]),
                phyparam["H_{}".format(filtnum)],
                phyparam["G1_{}".format(filtnum)],
                phyparam["G2_{}".format(filtnum)],
            )
        pdf.loc[cond, "preds"] = preds

    return pdf["i:magpsf_red"] - pdf["preds"]


@profile
def estimate_synodic_period(
    ssnamenr: str = None,
    pdf=None,
    phyparam=None,
    flavor="SHG1G2",
    Nterms_base=1,
    Nterms_band=1,
    period_range=(0.05, 1.2),
    sb_method="auto",
    return_extra_info=False,
    lt_correction=True,
):
    """Estimate the synodic period of a Solar System object seen by Fink

    Parameters
    ----------
    ssnamenr: str
        SSO number (we do not resolve name yet)
    pdf: pandas DataFrame, optional
        Pandas DataFrame with Fink SSO data for one object.
        If not specified, data will be downloaded from Fink servers
        using `ssnamenr`.
    phyparam: dict, optional
        Dictionary containing physical properties (phase curve, etc.)
        of the object. If not specified, they will be recomputed
        from the data.
    flavor: str, optional
        Model flavor: SHG1G2 (default), HG1G2, HG12, or HG
    Nterms_base: int, optional
        Number of frequency terms to use for the
        base model common to all bands. Default is 1.
    Nterms_band: int, optional
        Number of frequency terms to use for the
        residuals between the base model and
        each individual band. Default is 1.
    period_range: tupe of float, optional
        (min_period, max_period) for the search, in days.
        Default is (0.05, 1.2), that is between
        1.2 hours and 28.8 hours.
    sb_method: str, optional
        Specify the single-band lomb scargle implementation to use.
        See https://docs.astropy.org/en/stable/api/astropy.timeseries.LombScargleMultiband.html#astropy.timeseries.LombScargleMultiband.autopower
        If nifty-ls is installed, one can also specify fastnifty. Although
        in this case it does not work yet for Nterms_* higher than 1.
    return_extra_info: bool, optional
        If True, returns also the fitted model, and the original
        SSO data used for the fit. Default is False.
    lt_correction: bool, optional
        Apply light travel correction. Default is True.

    Returns
    -------
    best_period: float
        Best period found, in hour
    reduced_chi2: float
        Reduced chi2 (chi2 per ddof)
    model: object, optional
        If `return_extra_info`, the fitted model is returned
    pdf: pandas DataFrame, optional
        If `return_extra_info`, the original data is returned.
        Note that the extra field "residuals" contains the
        `observation - model` data, in magnitude.

    Examples
    --------
    >>> ssnamenr = 2363
    >>> P, chi2 = estimate_synodic_period(ssnamenr, flavor="SHG1G2", Nterms_base=2)
    >>> assert int(P) == 20, P

    >>> P_HG, chi2_HG = estimate_synodic_period(ssnamenr, flavor="HG", Nterms_base=2)
    >>> assert chi2 < chi2_HG, (chi2, chi2_HG)

    # by default we apply the light travel correction. Disable it.
    >>> P_no_lt, chi2_no_lt = estimate_synodic_period(ssnamenr, flavor="SHG1G2", Nterms_base=2, lt_correction=False)
    >>> assert chi2 < chi2_no_lt, (chi2, chi2_no_lt)


    One can also use the nifty-ls implementation (faster and more accurate)
    # TODO: check alias between astropy and nifty-ls...
    >>> P_nifty, _ = estimate_synodic_period(ssnamenr, flavor="SHG1G2", sb_method="fastnifty")
    >>> p1 = np.isclose(P, P_nifty, rtol=1e-1)
    >>> p2 = np.isclose(P, 2 * P_nifty, rtol=1e-1)
    >>> assert p1 or p2, (P, P_nifty)

    One can also directly specify the Pandas dataframe with Fink data:
    # TODO: check alias between astropy and nifty-ls...
    >>> r = requests.post("https://fink-portal.org/api/v1/sso", json={"n_or_d": ssnamenr, "withEphem": True, "output-format": "json"})
    >>> pdf = pd.read_json(io.BytesIO(r.content))
    >>> P_from_pdf, _ = estimate_synodic_period(pdf=pdf, flavor="SHG1G2")
    >>> p1 = np.isclose(P, P_from_pdf, rtol=1e-1)
    >>> p2 = np.isclose(P, 2 * P_from_pdf, rtol=1e-1)
    >>> assert p1 or p2, (P, P_from_pdf)
    """
    if pdf is None:
        if ssnamenr is not None:
            # TODO: use quaero
            r = requests.post(
                "https://fink-portal.org/api/v1/sso",
                json={"n_or_d": ssnamenr, "withEphem": True, "output-format": "json"},
            )
        else:
            _LOG.error("You need to specify either `ssnamenr` or `pdf`.")

        pdf = pd.read_json(io.BytesIO(r.content))

    if phyparam is None:
        # get the physical parameters with the latest data
        phyparam = extract_physical_parameters(pdf, flavor)

    # Compute the residuals (obs - model)
    residuals = compute_residuals(pdf, flavor, phyparam)

    if lt_correction:
        # Speed of light in AU/day
        time = compute_light_travel_correction(pdf["i:jd"], pdf["Dobs"])
    else:
        time = pdf["i:jd"]

    model = LombScargleMultiband(
        time,
        residuals,
        pdf["i:fid"],
        pdf["i:sigmapsf"],
        nterms_base=Nterms_base,
        nterms_band=Nterms_band,
    )

    frequency, power = model.autopower(
        method="fast",
        sb_method=sb_method,
        minimum_frequency=1 / period_range[1],
        maximum_frequency=1 / period_range[0],
    )
    freq_maxpower = frequency[np.argmax(power)]
    # Rotation in days (2* LS value: double-peaked lightcurve)
    # TODO: I need to be convinced...
    best_period = 2 / freq_maxpower

    out = model.model(time.to_numpy(), freq_maxpower)
    prediction = np.zeros_like(residuals)
    for index, filt in enumerate(pdf["i:fid"].unique()):
        if filt == 3:
            continue
        cond = pdf["i:fid"] == filt
        prediction[cond] = out[index][cond]

    chi2 = np.sum(((residuals - prediction) / pdf["i:sigmapsf"].to_numpy()) ** 2)
    reduced_chi2 = chi2 / len(residuals - 1)

    if return_extra_info:
        pdf["residuals"] = residuals
        return best_period * 24, reduced_chi2, frequency, power, model, pdf
    else:
        return best_period * 24, reduced_chi2


if __name__ == "__main__":
    """Execute the unit test suite"""

    # Run the Spark test suite
    regular_unit_tests(globals())



# Copyright 2022 AstroLab Software
# Author: Julien Peloton
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import pandas as pd
import numpy as np

from urllib.error import URLError

from fink_utils import __file__


def return_list_of_eg_host(full_simbad_conversion=False) -> list:
    """Return potential SN host names

    This includes:
    - List of object names in SIMBAD that would correspond to extra-galactic object
    - Unknown objects
    - objects with failed crossmatch

    In practice, this exclude galactic objects from SIMBAD.

    Parameters
    ----------
    full_simbad_conversion: bool
        If True, download the file containing the taxonomy change, and
        include old and new taxonomy in the output. If False, we only add
        manually the new labels we know have changed. In principle, we make sure
        that we do not miss ones, but if the taxonomy changed, the output could
        be incomplete. The former is in principle the most robust technics, but
        it is much slower (x1000) than the latter. Default is False.

    Returns
    -------
    out: list
        List of labels

    Examples
    --------
    >>> gals = return_list_of_eg_host(full_simbad_conversion=True)
    >>> print(len(gals))
    42

    >>> gals = return_list_of_eg_host(full_simbad_conversion=False)
    >>> print(len(gals))
    30
    """
    # old taxonomy
    list_simbad_galaxies = [
        "galaxy",
        "Galaxy",
        "EmG",
        "Seyfert",
        "Seyfert_1",
        "Seyfert_2",
        "BlueCompG",
        "StarburstG",
        "LSB_G",
        "HII_G",
        "High_z_G",
        "GinPair",
        "GinGroup",
        "BClG",
        "GinCl",
        "PartofG",
        "Compact_Gr_G",
        "IG",
        "PairG",
        "GroupG",
        "ClG",
        "SuperClG",
        "Void",
    ]

    cds = [
        "Unknown",
        "Candidate_SN*",
        "SN",
        "Transient",
        "Fail",
        "Fail 504",
    ] + list_simbad_galaxies

    if full_simbad_conversion:
        conv = get_conversion_dic()
        cds_with_new_taxonomy = [old2new(conv, i) for i in cds]
    else:
        # fields that really changed
        cds_with_new_taxonomy = ["SN*_Candidate"]

    out = np.concatenate((cds, cds_with_new_taxonomy))

    return np.unique(out)


def get_conversion_dic(path: str = None, remove_unknown: bool = True) -> pd.DataFrame:
    """Read the file containing the mapping between old and new otypes

    Parameters
    ----------
    path: str
        Path to the file. Can be an URL:
        https://simbad.cds.unistra.fr/guide/otypes.labels.txt
    remove_unknown: bool
        If True, remove the row containing the class `Unknown` from
        the DataFrame. Default is True.

    Returns
    -------
    pdf: pd.DataFrame
        Data formatted in a pandas DataFrame: otype, old_label, new_label

    Examples
    --------
    >>> pdf = get_conversion_dic(remove_unknown=False)
    >>> print(len(pdf))
    199

    >>> pdf = get_conversion_dic()
    >>> print(len(pdf))
    198
    """
    if path is None:
        path = "https://simbad.cds.unistra.fr/guide/otypes.labels.txt"

    try:
        pdf = pd.read_csv(
            path,
            sep="|",
            skiprows=[0, 1, 3],
            skipfooter=2,
            dtype="str",
            header=0,
            engine="python",
        )
    except URLError:
        curdir = os.path.dirname(os.path.abspath(__file__))
        fn = curdir + "/xmatch/otypes.txt"
        pdf = pd.read_csv(
            fn,
            sep="|",
            skiprows=[0, 1, 3],
            skipfooter=2,
            dtype="str",
            header=0,
            engine="python",
        )

    pdf = pdf.rename(columns={i: i.strip() for i in pdf.columns})

    pdf = pdf[["otype", "old_label", "new_label"]]

    pdf = pdf.applymap(lambda x: x.strip())

    if remove_unknown:
        pdf = pdf[pdf["old_label"] != "Unknown"]

    return pdf


def old2new(conv, old_label=""):
    """Return the new label name in SIMBAD from an old label

    Parameters
    ----------
    conv: pd.DataFrame
        DataFrame containing all the labels (see `get_conversion_dic`)
    old_label: str
        Old label in SIMBAD

    Returns
    -------
    new_label: str
        New label in SIMBAD corresponding to the old label

    Examples
    --------
    >>> path = 'https://simbad.cds.unistra.fr/guide/otypes.labels.txt'
    >>> pdf = get_conversion_dic(path)
    >>> new_label = old2new(pdf, 'BlueCompG')
    >>> print(new_label)
    BlueCompactG
    """
    out = conv[conv["old_label"] == old_label]["new_label"].to_numpy()

    if len(out) == 1:
        return out[0]
    elif len(out) == 0:
        return ""

    raise ValueError("Label conversion is ambiguous: {} --> {}".format(old_label, out))


def new2old(conv, new_label=""):
    """Return the old label name in SIMBAD from a new label

    Parameters
    ----------
    conv: pd.DataFrame
        DataFrame containing all the labels (see `get_conversion_dic`)
    new_label: str
        New label in SIMBAD

    Returns
    -------
    old_label: str
        Old label in SIMBAD corresponding to the new label

    Examples
    --------
    >>> path = 'https://simbad.cds.unistra.fr/guide/otypes.labels.txt'
    >>> pdf = get_conversion_dic(path)
    >>> old_label = new2old(pdf, 'GtowardsGroup')
    >>> print(old_label)
    GinGroup
    """
    out = conv[conv["new_label"] == new_label]["old_label"].to_numpy()

    if len(out) == 1:
        return out[0]
    elif len(out) == 0:
        return ""

    raise ValueError("Label conversion is ambiguous: {} --> {}".format(new_label, out))


def get_simbad_labels(which: str, remove_unknown=True):
    """Get list of labels in SIMBAD.

    Parameters
    ----------
    which: str
        Choose between: `old`, `new`, `old_and_new`, `otype`.
        `old_and_new` is the unique concatenation of old and new labels.
    remove_unknown: bool
        If True, remove the row containing the class `Unknown` from
        the DataFrame. Default is True.

    Returns
    -------
    out: list
        List of labels (str)


    Examples
    --------
    >>> labels = get_simbad_labels(which='old')
    >>> print(len(labels))
    198

    >>> assert 'Candidate_YSO' in labels

    >>> labels = get_simbad_labels(which='new')
    >>> print(len(labels))
    198

    >>> assert 'YSO_Candidate' in labels

    >>> labels = get_simbad_labels(which='old_and_new')
    >>> print(len(labels))
    315

    >>> assert 'Candidate_YSO' in labels
    >>> assert 'YSO_Candidate' in labels

    >>> labels = get_simbad_labels(which='otype')
    >>> print(len(labels))
    198

    >>> assert 'Y*?' in labels
    """
    pdf = get_conversion_dic(remove_unknown=remove_unknown)
    if which == "old":
        return pdf["old_label"].to_numpy()
    elif which == "new":
        return pdf["new_label"].to_numpy()
    elif which == "old_and_new":
        old = pdf["old_label"].to_numpy()
        new = pdf["new_label"].to_numpy()
        out = np.concatenate((old, new))
        return np.unique(out)
    elif which == "otype":
        return pdf["otype"].to_numpy()


if __name__ == "__main__":
    """Execute the test suite"""
    import sys
    import doctest

    sys.exit(doctest.testmod()[0])



# Copyright 2019-2024 AstroLab Software
# Author: Roman Le Montagner
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
from typing import Tuple


def vect_mag2fluxcal_snana(
    magpsf: "np.array[np.float]", sigmapsf: "np.array[np.float]"
) -> Tuple["np.array[np.float]", "np.array[np.float]"]:
    """Conversion from magnitude to Fluxcal from SNANA manual.

    vectorized version

    Parameters
    ----------
    magpsf: float
        PSF-fit magnitude from ZTF
    sigmapsf: float

    Returns
    -------
    fluxcal: float
        Flux cal as used by SNANA
    fluxcal_err: float
        Absolute error on fluxcal (the derivative has a minus sign)

    """
    magpsf = np.where(np.equal(magpsf, None), np.nan, magpsf)
    sigmapsf = np.where(np.equal(sigmapsf, None), np.nan, sigmapsf)

    fluxcal = 10 ** (-0.4 * magpsf) * 10 ** (11)
    fluxcal_err = 9.21034 * 10**10 * np.exp(-0.921034 * magpsf) * sigmapsf

    return fluxcal, fluxcal_err


def vect_apparent_flux(
    magpsf: "np.array[np.float]",
    sigmapsf: "np.array[np.float]",
    magnr: "np.array[np.float]",
    sigmagnr: "np.array[np.float]",
    isdiffpos: "np.array[str]",
    jansky: bool = True,
) -> Tuple["np.array[np.float]", "np.array[np.float]"]:
    """Compute apparent flux from difference magnitude supplied by ZTF

    Implemented according to p.107 of the ZTF Science Data System Explanatory Supplement
    https://irsa.ipac.caltech.edu/data/ZTF/docs/ztf_explanatory_supplement.pdf
    vectorized version

    Parameters
    ----------
    magpsf,sigmapsf; floats
        magnitude from PSF-fit photometry, and 1-sigma error
    magnr,sigmagnr: floats
        magnitude of nearest source in reference image PSF-catalog
        within 30 arcsec and 1-sigma error
    isdiffpos: str
        t or 1 => candidate is from positive (sci minus ref) subtraction;
        f or 0 => candidate is from negative (ref minus sci) subtraction
    jansky: bool
        If True, normalise units to Jansky. Default is True.

    Returns
    -------
    dc_flux: float
        Apparent flux
    dc_sigflux: float
        Error on apparent flux
    """
    magpsf = np.where(np.equal(magpsf, None), np.nan, magpsf)
    magnr = np.where(magnr < 0.0, np.nan, magnr)

    difference_flux = 10 ** (-0.4 * magpsf)
    difference_sigflux = (sigmapsf / 1.0857) * difference_flux

    ref_flux = 10 ** (-0.4 * magnr)
    ref_sigflux = (sigmagnr / 1.0857) * ref_flux

    # add or subract difference flux based on isdiffpos
    dc_flux = np.where(
        (isdiffpos == "t") | (isdiffpos == "1"),
        ref_flux + difference_flux,
        ref_flux - difference_flux,
    )

    # assumes errors are independent. Maybe too conservative.
    dc_sigflux = np.sqrt(difference_sigflux**2 + ref_sigflux**2)

    if jansky:
        dc_flux *= 3631
        dc_sigflux *= 3631

    return dc_flux, dc_sigflux


def vect_dc_mag(
    magpsf: "np.array[np.float]",
    sigmapsf: "np.array[np.float]",
    magnr: "np.array[np.float]",
    sigmagnr: "np.array[np.float]",
    isdiffpos: "np.array[str]",
) -> Tuple["np.array[np.float]", "np.array[np.float]"]:
    """Compute apparent magnitude from difference magnitude supplied by ZTF

    Implemented according to p.107 of the ZTF Science Data System Explanatory Supplement
    https://irsa.ipac.caltech.edu/data/ZTF/docs/ztf_explanatory_supplement.pdf
    vectorized version

    Parameters
    ----------
    magpsf,sigmapsf
        magnitude from PSF-fit photometry, and 1-sigma error
    magnr,sigmagnr
        magnitude of nearest source in reference image PSF-catalog
        within 30 arcsec and 1-sigma error
    isdiffpos
        t or 1 => candidate is from positive (sci minus ref) subtraction
        f or 0 => candidate is from negative (ref minus sci) subtraction

    Returns
    -------
    dc_mag: float
        Apparent magnitude
    dc_sigmag: float
        Error on apparent magnitude
    """
    dc_flux, dc_sigflux = vect_apparent_flux(
        magpsf, sigmapsf, magnr, sigmagnr, isdiffpos, jansky=False
    )

    # apparent mag and its error from fluxes
    dc_mag = -2.5 * np.log10(dc_flux)

    dc_sigmag = dc_sigflux / dc_flux * 1.0857

    return dc_mag, dc_sigmag



# Copyright 2019-2022 AstroLab Software
# Author: Roman Le Montagner
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest
import requests
import pandas as pd
import numpy as np
import time as t

from fink_utils.photometry.conversion import dc_mag
from fink_utils.photometry.conversion import mag2fluxcal_snana

from fink_utils.photometry.vect_conversion import vect_mag2fluxcal_snana
from fink_utils.photometry.vect_conversion import vect_dc_mag


@pytest.fixture
def requests_alerts():
    r = requests.post(
        "https://fink-portal.org/api/v1/latests",
        json={"class": "Seyfert_2", "n": "500", "columns": "i:objectId"},
    )

    # Format output in a DataFrame
    pdf = pd.DataFrame.from_dict(r.json())

    request_object = ",".join(list(pdf["i:objectId"]))
    request_object += ",ZTF21abeyuqh,ZTF21abeyutr,ZTF21abeyuua"
    request_columns = (
        "i:objectId,i:candid,i:magpsf,i:sigmapsf,i:magnr,i:sigmagnr,i:isdiffpos"
    )

    r = requests.post(
        "https://fink-portal.org/api/v1/objects",
        json={
            "objectId": request_object,
            "output-format": "json",
            "columns": request_columns,
        },
    )

    # Format output in a DataFrame
    pdf = pd.DataFrame.from_dict(r.json())

    return pdf


def test_dc_mag(requests_alerts):
    t_before = t.time()
    res_dc_mag, res_sig_mag = vect_dc_mag(
        requests_alerts["i:magpsf"].values,
        requests_alerts["i:sigmapsf"].values,
        requests_alerts["i:magnr"].values,
        requests_alerts["i:sigmagnr"].values,
        requests_alerts["i:isdiffpos"].values,
    )
    t_vect = t.time() - t_before

    t_before = t.time()
    _dc_mag = requests_alerts.apply(
        lambda x: dc_mag(
            x["i:magpsf"],
            x["i:sigmapsf"],
            x["i:magnr"],
            x["i:sigmagnr"],
            x["i:isdiffpos"],
        ),
        axis=1,
        result_type="expand",
    )
    t_loop = t.time() - t_before

    t1 = np.allclose(res_dc_mag, _dc_mag[0])

    t2 = np.allclose(res_sig_mag, _dc_mag[1])

    assert np.all(np.allclose(t1, t2))
    assert t_vect < t_loop


def test_flux_snana(requests_alerts):
    t_before = t.time()
    vect_flux, vect_sigflux = vect_mag2fluxcal_snana(
        requests_alerts["i:magpsf"].values, requests_alerts["i:sigmapsf"].values
    )
    t_vect = t.time() - t_before

    t_before = t.time()
    _flux_snana = requests_alerts.apply(
        lambda x: mag2fluxcal_snana(x["i:magpsf"], x["i:sigmapsf"]),
        axis=1,
        result_type="expand",
    )
    t_loop = t.time() - t_before

    t1 = np.allclose(vect_flux, _flux_snana[0])

    t2 = np.allclose(vect_sigflux, _flux_snana[1])

    assert np.all(np.allclose(t1, t2))
    assert t_vect < t_loop



import logging
import datetime
import pytz

import fink_utils


class CustomTZFormatter(logging.Formatter):  # pragma: no cover
    """override logging.Formatter to use an aware datetime object"""

    def converter(self, timestamp):
        dt = datetime.datetime.fromtimestamp(timestamp)
        tzinfo = pytz.timezone("Europe/Paris")
        return tzinfo.localize(dt)

    def formatTime(self, record, datefmt=None):  # noqa: N802
        dt = self.converter(record.created)
        if datefmt:
            s = dt.strftime(datefmt)
        else:
            try:
                s = dt.isoformat(timespec="milliseconds")
            except TypeError:
                s = dt.isoformat()
        return s


class LoggerNewLine(logging.Logger):
    """
    A custom logger class adding only a method to print a newline.

    Examples
    --------
    logger.newline()
    """

    def __init__(self, name: str, level: int = 0) -> None:
        super().__init__(name, level)
        ch = logging.StreamHandler()

        self.setLevel(logging.DEBUG)

        # create console handler and set level to debug
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)

        # create formatter
        formatter = CustomTZFormatter(
            "%(asctime)s - %(name)s - %(levelname)s \n\t message: %(message)s"
        )

        # add formatter to ch
        ch.setFormatter(formatter)

        # add ch to logger
        self.addHandler(ch)

        blank_handler = logging.StreamHandler()
        blank_handler.setLevel(logging.DEBUG)
        blank_handler.setFormatter(logging.Formatter(fmt=""))
        self.console_handler = ch
        self.blank_handler = blank_handler

    def newline(self, how_many_lines=1):
        """
        Print blank line using the logger class

        Parameters
        ----------
        how_many_lines : int, optional
            how many blank line to print, by default 1
        """
        # Switch handler, output a blank line
        self.removeHandler(self.console_handler)
        self.addHandler(self.blank_handler)
        for _ in range(how_many_lines):
            self.info("\n")

        # Switch back
        self.removeHandler(self.blank_handler)
        self.addHandler(self.console_handler)


def init_logging(logger_name=fink_utils.__name__) -> LoggerNewLine:
    """
    Initialise a logger for the gcn stream

    Parameters
    ----------
    None

    Returns
    -------
    logger : Logger object
        A logger object for the logging management.

    Examples
    --------
    >>> l = init_logging()
    >>> type(l)
    <class 'logging.Logger'>
    """
    # create logger

    logging.setLoggerClass(LoggerNewLine)
    logger = logging.getLogger(logger_name)
    return logger



# Copyright 2019 AstroLab Software
# Author: Abhishek Chauhan, Julien Peloton
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
import glob
import shutil
import time

from fink_utils.broker.avroUtils import readschemafromavrofile
from fink_utils.broker.sparkUtils import to_avro, from_avro

from pyspark.sql import DataFrame
from pyspark.sql.functions import struct, lit
from pyspark.sql.avro.functions import to_avro as to_avro_native

# from fink_utils.test.tester import spark_unit_tests


def get_kafka_df(
    df: DataFrame,
    schema_path: str,
    fink_broker_version: str,
    fink_science_version: str,
    saveschema: bool = False,
    elasticc: bool = False,
) -> DataFrame:
    """Create and return a df to pubish to Kafka

    For a kafka output the dataframe should have the following columns:
    key: (optional) Using a unique key can prevent reading duplicate data
        as Kafka supports "at least once" write semantics
        and might result in duplicate writing
    value: (required)
    topic: (*optional) if a topic field is not present in the dataframe
        it has to be given while writing to kafka
    This routine groups the DataFrame columns to be published to Kafka into
    a StructType column and convert it into avro(binary).
    To be able to decode the sent Kafka messages, the avro schema is stored
    at schema_path which is passed as an argument.

    Parameters
    ----------
    df: DataFrame
        A Spark DataFrame created after reading the science database (HBase)
    schema_path: str
        Path where to store the avro schema required for decoding the
        Kafka messages.
    saveschema: bool
        If True, save the alert schema on disk. Work only in Spark local mode,
        and for testing purposes. Default is False.

    Returns
    -------
    df: DataFrame
        A Spark DataFrame with an avro(binary) encoded Column named "value"
    """
    # Remove the status column before distribution
    cols = df.columns
    if "status" in cols:
        cols.remove("status")

    df = df.select(cols)

    # Create a StructType column in the df for distribution.
    # The contents and schema of the df can change over time
    df_struct = df.select(struct(df.columns).alias("struct"))

    # Convert into avro and save the schema
    if elasticc:
        # The idea is to force the output schema
        # Need better handling of this though...
        jsonschema = open(
            "/home/julien.peloton/elasticc/alert_schema/elasticc.v0_9.brokerClassification.avsc",
            "r",
        ).read()
        df_kafka = df_struct.select(to_avro_native("struct", jsonschema).alias("value"))
    else:
        df_kafka = df_struct.select(to_avro("struct").alias("value"))

    # Add a key based on schema versions
    df_kafka = df_kafka.withColumn(
        "key", lit("{}_{}".format(fink_broker_version, fink_science_version))
    )

    if saveschema:
        # Harcoded path that corresponds to the schema used
        # for alert redistribution.
        schema_path = "schemas/distribution_schema_new.avsc"

        # Do not work on a DFS like HDFS obviously.
        # Only local mode & for testing purposes
        toto = df.writeStream.foreachBatch(
            lambda x, y: save_avro_schema_stream(x, y, schema_path, False)
        ).start()
        time.sleep(10)

        # Note that the entire Spark application will stop.
        toto.stop()

    return df_kafka


def save_avro_schema_stream(
    df: DataFrame, epochid: int, schema_path=None, save_on_hdfs: bool = False
):
    """Extract schema from an alert of the stream, and save it on disk.

    Mostly for debugging purposes - do not work in cluster mode (local only).
    Typically:
    schema_path = ...
    toto = df_stream.writeStream.foreachBatch(
        lambda x, y: save_avro_schema_stream(x, y, schema_path)
    ).start()
    time.sleep(10)
    toto.stop()

    Parameters
    ----------
    df: DataFrame
        Micro-batch Dataframe containing alerts
    epochid: int
        Offset of the micro-batch
    """
    save_avro_schema(df, schema_path, save_on_hdfs)


def save_avro_schema(df: DataFrame, schema_path: str, save_on_hdfs: bool = False):
    """Writes the avro schema to a file at schema_path

    This routine checks if an avro schema file exist at the given path
    and creates one if it doesn't.
    To automatically change the schema with changing requirements, ensure to
    delete the schema file at the given path whenever the structure of DF
    read from science db or the contents to be distributed are changed.

    Parameters
    ----------
    df: DataFrame
        A Spark DataFrame
    schema_path: str
        Path where to store the avro schema
    """
    # Check if the file exists
    if not os.path.isfile(schema_path):
        # Store the df as an avro file
        if save_on_hdfs:
            prefix_path = "hdfs:///"
        else:
            prefix_path = "file:///"
        path_for_avro = os.path.join(
            prefix_path, os.environ["PWD"][1:], "flatten_hbase.avro"
        )
        if os.path.exists(path_for_avro):
            shutil.rmtree(path_for_avro)
        df.write.format("avro").save(path_for_avro)

        # Read the avro schema from .avro file
        if save_on_hdfs:
            avro_file = glob.glob(path_for_avro + "/.part*")[0]
        else:
            avro_file = glob.glob(
                os.path.join(os.environ["PWD"], "flatten_hbase.avro", ".part*")
            )[0]
        avro_schema = readschemafromavrofile(avro_file)

        # Write the schema to a file for decoding Kafka messages
        with open(schema_path, "w") as f:
            json.dump(avro_schema, f, indent=2)

        # Remove .avro files and directory
        shutil.rmtree(path_for_avro)
    else:
        msg = """
            {} already exists - cannot write the new schema
        """.format(schema_path)
        print(msg)


def decode_kafka_df(df_kafka: DataFrame, schema_path: str) -> DataFrame:
    """Decode the DataFrame read from Kafka

    The DataFrame read from Kafka contains the following columns:
    key: binary
    value: binary
    topic: string
    partition: int
    offset: long
    timestamp: long
    timestampType: integer
    The value column contains the structured data of the alert encoded into
    avro(binary). This routine creates a Spark DataFrame with a decoded
    StructType column using the avro schema at schema_path.

    Parameters
    ----------
    df_kafka: DataFrame
        A Spark DataFrame created after reading the Kafka Source
    schema_path: str
        Path where the avro schema to decode the Kafka message is stored

    Returns
    -------
    df: DataFrame
        A Spark DataFrame with a StructType Column with decoded data of
        the avro(binary) column named "value"

    Examples
    --------
    # >>> from pyspark.sql.functions import col
    # >>> df = spark.sparkContext.parallelize(zip(
    # ...     ["ZTF18aceatkx", "ZTF18acsbjvw"],
    # ...     [697251923115015002, 697251921215010004],
    # ...     [20.393772, 20.4233877],
    # ...     [-25.4669463, -27.0588511],
    # ...     ["Star", "Unknown"])).toDF([
    # ...       "objectId", "candid", "candidate_ra",
    # ...       "candidate_dec", "cross_match_alerts_per_batch"])
    # >>> df.show()
    # +------------+------------------+------------+-------------+----------------------------+
    # |    objectId|            candid|candidate_ra|candidate_dec|cross_match_alerts_per_batch|
    # +------------+------------------+------------+-------------+----------------------------+
    # |ZTF18aceatkx|697251923115015002|   20.393772|  -25.4669463|                        Star|
    # |ZTF18acsbjvw|697251921215010004|  20.4233877|  -27.0588511|                     Unknown|
    # +------------+------------------+------------+-------------+----------------------------+
    # <BLANKLINE>
    # >>> temp_schema = os.path.join("file:///", os.environ["PWD"][1:], "temp_schema")
    # >>> save_avro_schema(df, temp_schema)
    # >>> df_kafka = get_kafka_df(df, '', "2.2", "2.0.0") # Encode the data into avro
    # >>> df_decoded = decode_kafka_df(df_kafka, temp_schema) # Decode the avro df
    # >>> df_decoded.printSchema()
    # root
    #  |-- struct: struct (nullable = true)
    #  |    |-- objectId: string (nullable = true)
    #  |    |-- candid: long (nullable = true)
    #  |    |-- candidate_ra: double (nullable = true)
    #  |    |-- candidate_dec: double (nullable = true)
    #  |    |-- cross_match_alerts_per_batch: string (nullable = true)
    # <BLANKLINE>
    # >>> df_decoded.select(col("struct.*")).show()
    # +------------+------------------+------------+-------------+----------------------------+
    # |    objectId|            candid|candidate_ra|candidate_dec|cross_match_alerts_per_batch|
    # +------------+------------------+------------+-------------+----------------------------+
    # |ZTF18aceatkx|697251923115015002|   20.393772|  -25.4669463|                        Star|
    # |ZTF18acsbjvw|697251921215010004|  20.4233877|  -27.0588511|                     Unknown|
    # +------------+------------------+------------+-------------+----------------------------+
    # <BLANKLINE>
    # >>> os.remove(temp_schema)
    """
    # Read the avro schema
    with open(schema_path) as f:
        avro_schema = json.dumps(json.load(f))

    # Decode the avro(binary) column
    df = df_kafka.select(from_avro("value", avro_schema).alias("struct"))

    return df


def get_distribution_offset(
    offsetfile: str, startingoffset_dist: str = "latest"
) -> int:
    """Read and return distribution offset from file

    Parameters
    ----------
    offsetfile: str
        the path of offset file for distribution
    startingoffset_dist: str, optional
        Offset(timestamp) from where to start the distribution. Options are
        latest (read timestamp from file), earliest (from the beginning of time),
        timestamp (custom timestamp input by user)

    Returns
    -------
    timestamp: int
        a timestamp (typical unix timestamp: time in ms since epoch)

    Examples
    --------
    # set a test timestamp
    >>> test_t = int(round(time.time() * 1000))
    >>> with open('dist.offset.test', 'w') as f: # write to a file
    ...     string = "distributed till, {}".format(test_t)
    ...     f.write(string)
    ...
    31
    >>> min_timestamp = get_distribution_offset("invalidFile", "latest") # test 1 (wrong file name)
    >>> min_timestamp
    100
    >>> min_timestamp = get_distribution_offset("dist.offset.test", "earliest") # test 2 (given earliest)
    >>> min_timestamp
    100
    >>> min_timestamp = get_distribution_offset("dist.offset.test", "latest") # test 3 (given latest)
    >>> min_timestamp == test_t
    True
    >>> min_timestamp = get_distribution_offset("dist.offset.test") # test 4 (no offset given)
    >>> min_timestamp == test_t
    True
    >>> custom_t = int(round(time.time() * 1000)) # test 5 (custom timestamp given)
    >>> min_timestamp = get_distribution_offset("dist.offset.test", custom_t)
    >>> min_timestamp == custom_t
    True
    >>> os.remove('dist.offset.test') # Delete offset file
    """
    # if the offset file doesn't exist or is empty or
    # one wants the earliest offset
    fileexist = os.path.isfile(offsetfile)
    if fileexist:
        negatifoffset = os.path.getsize(offsetfile) <= 0
    else:
        negatifoffset = False
    if not fileexist or negatifoffset or startingoffset_dist == "earliest":
        # set a default
        min_timestamp = 100
    else:
        if startingoffset_dist == "latest":
            with open(offsetfile, "r") as f:
                line = f.readlines()[-1]
                min_timestamp = int(line.split(", ")[-1])
        else:
            # user given timestamp
            min_timestamp = int(startingoffset_dist)

    return min_timestamp


def group_df_into_struct(df: DataFrame, colfamily: str, key: str) -> DataFrame:
    """Group columns of a df into a struct column

    Notes
    -----
    Currently, the dataframe is transformed by splitting it into
    two dataframes, reshaping one of them and then using a join.
    This might consume more resources than necessary and should be
    optimized in the future if required.
    If we have a df with the following schema:
    root
     |-- objectId: string (nullable = true)
     |-- candidate_ra: double (nullable = true)
     |-- candidate_dec: double (nullable = true)
    and we want to group all `candidate_*` into a struct like:
    root
     |-- objectId: string (nullable = true)
     |-- candidate: struct (nullable = false)
     |    |-- ra: double (nullable = true)
     |    |-- dec: double (nullable = true)

    Parameters
    ----------
    df: Spark DataFrame
        a Spark dataframe with flat columns
    colfamily: str
        prefix of columns to be grouped into a struct
    key: str
        a column with unique values (used for join)

    Returns
    -------
    df: Spark DataFrame
        a Spark dataframe with columns grouped into struct

    Examples
    --------
    >>> df = spark.sparkContext.parallelize(zip(
    ...     ["ZTF18aceatkx", "ZTF18acsbjvw"],
    ...     [697251923115015002, 697251921215010004],
    ...     [20.393772, 20.4233877],
    ...     [-25.4669463, -27.0588511],
    ...     ["Star", "Unknown"])).toDF([
    ...       "objectId", "candid", "candidate_ra",
    ...       "candidate_dec", "cross_match_alerts_per_batch"])
    >>> df.printSchema()
    root
     |-- objectId: string (nullable = true)
     |-- candid: long (nullable = true)
     |-- candidate_ra: double (nullable = true)
     |-- candidate_dec: double (nullable = true)
     |-- cross_match_alerts_per_batch: string (nullable = true)
    <BLANKLINE>
    >>> df = group_df_into_struct(df, 'candidate', 'objectId')
    >>> df.printSchema()
    root
     |-- objectId: string (nullable = true)
     |-- candid: long (nullable = true)
     |-- cross_match_alerts_per_batch: string (nullable = true)
     |-- candidate: struct (nullable = false)
     |    |-- ra: double (nullable = true)
     |    |-- dec: double (nullable = true)
    <BLANKLINE>
    """
    struct_cols = []
    flat_cols = []

    pos = len(colfamily) + 1

    for col_ in df.columns:
        if col_.startswith(colfamily + "_"):
            struct_cols.append(col_)
        else:
            flat_cols.append(col_)

    # dataframe with columns other than 'columnFamily_*'
    df1 = df.select(flat_cols)

    new_col_names = []
    new_col_names.append(key)

    # dataframe with key + 'columnFamily_*'
    df2 = df.select(new_col_names + struct_cols)

    struct_cols = [x[pos:] for x in struct_cols]

    new_col_names.extend(struct_cols)
    df2_renamed = df2.toDF(*new_col_names)

    # Group 'columnFamily_*' into a struct
    df2_struct = df2_renamed.select(key, struct(*struct_cols).alias(colfamily))

    # join the two dataframes based on 'key'
    df_new = df1.join(df2_struct, key)

    return df_new


def write_to_kafka(
    sdf,
    key,
    kafka_bootstrap_servers,
    kafka_sasl_username,
    kafka_sasl_password,
    topic_name,
    checkpoint_stream_path,
    processingTime,
):
    """
    Send data to a Kafka cluster using Apache Spark

    Parameters
    ----------
    sdf: Spark DataFrame
        DataFrame
    key: str
        key for each Avro message
    kafka_bootstrap_servers: str
        Comma-separated list of ip:port of the Kafka machines
    kafka_sasl_username: str
        Username for writing into the Kafka cluster
    kafka_sasl_password: str
        Password for writing into the Kafka cluster
    topic_name: str
        Kafka topic (does not need to exist)
    checkpoint_stream_path: str
        path where to store the stream checkpoint files
    processingTime: float
        processing interval time between batch
    Return
    ------
    disquery: Spark Streaming DataFrame
        Streaming DataFrame
    """
    # Create a StructType column in the df for distribution.
    df_struct = sdf.select(struct(sdf.columns).alias("struct"))
    df_kafka = df_struct.select(to_avro("struct").alias("value"))
    df_kafka = df_kafka.withColumn("key", key)

    # Send schema
    disquery = (
        df_kafka.writeStream.format("kafka")
        .option("kafka.bootstrap.servers", kafka_bootstrap_servers)
        .option("kafka.sasl.username", kafka_sasl_username)
        .option("kafka.sasl.password", kafka_sasl_password)
        .option("topic", topic_name)
        .option("checkpointLocation", checkpoint_stream_path)
        .trigger(processingTime="{} seconds".format(processingTime))
        .start()
    )

    return disquery



# Copyright 2020-2024 AstroLab Software
# Author: Julien Peloton
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
import pandas as pd

import pyspark.sql.functions as F
from pyspark.sql.functions import pandas_udf, PandasUDFType
from pyspark.sql.types import (
    ArrayType,
    BooleanType,
    StructType,
    MapType,
    StringType,
    FloatType,
)
from pyspark.sql import DataFrame

import importlib
import logging

_LOG = logging.getLogger(__name__)


def concat_col(
    df,
    colname: str,
    prefix: str = "c",
    current: str = "candidate",
    history: str = "prv_candidates",
):
    """Concatenate historical and current measurements

    Add new column to the DataFrame named `prefix`+`colname`, containing
    the concatenation of historical and current measurements.

    Parameters
    ----------
    df: DataFrame
        Pyspark DataFrame containing alert data
    colname: str
        Name of the column to add (without the prefix)
    prefix: str
        Additional prefix to add to the column name. Default is 'c'.
    current: str
        Name of the field containing current `colname` measurement, to extract
        `current.colname`. Usually a struct type field. Default is `candidate`
        from ZTF schema.
    history: str
        Name of the field containing history for `colname` measurements,
        to extract `history.colname`. Usually a list of struct type field.
        Default is `prv_candidates` from ZTF schema.

    Returns
    -------
    df: DataFrame
        Dataframe with new column containing the concatenation of
        historical and current measurements.
    """
    return df.withColumn(
        prefix + colname,
        F.when(
            df["{}.{}".format(history, colname)].isNotNull(),
            F.concat(
                df["{}.{}".format(history, colname)],
                F.array(df["{}.{}".format(current, colname)]),
            ),
        ).otherwise(F.array(df["{}.{}".format(current, colname)])),
    )


def extract_values(cmagpsf, cdiffmaglim, onlyfainterlimits=False):
    """Extract the first upper values before measurements start

    Parameters
    ----------
    cmagpsf: np.array
        Array of magpsf (history + current)
    cdiffmaglim: np.array
        Array of diffmaglim (history + current)
    onlyfainterlimits: bool, optional
        If True, only consider upper limits that are fainter than the
        first measurement. Default False

    Returns
    -------
    val: float
        Last diffmaglim values before measurements start

    Examples
    --------
    >>> cmagpsf = [np.nan, np.nan, 1.0, 2.0, np.nan, 3.0]
    >>> cdiffmaglim = [0.3, 0.2, 0.4, 0.7, 0.4, 0.2]
    >>> val = extract_values(cmagpsf, cdiffmaglim)
    >>> assert val == 0.2
    """
    if len(cmagpsf) == 0:
        return np.nan

    if not np.isnan(cmagpsf[0]):
        # young transient must start with NaN
        return np.nan

    if np.alltrue(np.isnan(cmagpsf)):
        # all NaNs
        return np.nan

    # 0..N_alert
    array_indices = np.arange(len(cmagpsf))

    # Position of the last upper values
    first_mag = np.nan
    pos = 0
    while np.isnan(first_mag) and pos < len(cmagpsf):
        first_mag = cmagpsf[pos]
        pos += 1

    if (pos == len(cmagpsf)) and np.isnan(first_mag):
        return np.nan

    # Arrays with only upper values
    diffmaglim = cdiffmaglim[array_indices <= pos - 2]

    # Selecting only fainter limits if required
    if onlyfainterlimits:
        diffmaglim = diffmaglim[diffmaglim > first_mag]

    if len(diffmaglim) == 0:
        return np.nan
    else:
        return diffmaglim[-1]


@pandas_udf(MapType(StringType(), ArrayType(FloatType())), PandasUDFType.SCALAR)
def extend_lc_with_upper_limits(
    cmagpsf, csigmapsf, cfid, cdiffmaglim, onlyfainterlimits=False
):
    """Extend valid measurements with the last upper limit for each band

    Notes
    -----
    The extended time-series can have:
    1. new measurements in r & g
    2. new measurements in r or g
    3. no new measurements

    In the case of (2) or (3), missing addition means the history
    already starts with a valid measurement, and there is nothing to
    be further added.

    Parameters
    ----------
    cmagpsf: pd.Series of np.array
        Series of arrays of magpsf (history + current)
    csigmapsf: pd.Series of np.array
        Series of arrays of sigmapsf (history + current)
    cfid: pd.Series of np.array
        Series of arrays of fid (history + current)
    cdiffmaglim: pd.Series of np.array
        Series of arrays of diffmaglim (history + current)
    onlyfainterlimits: bool, optional
        If True, only consider upper limits that are fainter than the
        first measurement. Default False

    Returns
    -------
    pd.Series
        new column with dictionaries containing extended
        time-series (`cmagpsf_ext` & `csigmapsf_ext`). Note that
        added values for sigmapsf is always 0.2 mag.

    Examples
    --------
    >>> to_expand = ["jd", "fid", "magpsf", "sigmapsf", "diffmaglim", "sigmapsf"]
    >>> prefix = "c"
    >>> for colname in to_expand:
    ...     df = concat_col(df, colname, prefix=prefix)

    >>> df = spark.read.format("parquet").load("datatest")
    >>> df = df.withColumn("ext", extend_lc_with_upper_limits(
    ...     "cmagpsf", "csigmapsf", "cfid", "cdiffmaglim"))
    >>> df = df.withColumn("cmagpsf_ext", df["ext"].getItem("cmagpsf_ext"))
    >>> df = df.withColumn("csigmapsf_ext", df["ext"].getItem("csigmapsf_ext"))

    >>> pdf = df.select(["cmagpsf", "cmagpsf_ext"]).toPandas()

    >>> cmagpsf, cmagpsf_ext = pdf.to_numpy()[0]
    >>> assert len(cmagpsf) == len(cmagpsf_ext) # they have the same size

    >>> assert len(cmagpsf[~np.isnan(cmagpsf)]) <= len(cmagpsf_ext[~np.isnan(cmagpsf_ext)])
    """
    if not isinstance(onlyfainterlimits, bool):
        # Series to boolean
        onlyfainterlimits = True
    out = []
    for index in range(len(cmagpsf)):
        row = {"cmagpsf_ext": cmagpsf[index], "csigmapsf_ext": csigmapsf[index]}
        for fid in [1, 2]:
            mask = cfid[index] == fid
            val = extract_values(
                cmagpsf[index][mask],
                cdiffmaglim[index][mask],
                onlyfainterlimits=onlyfainterlimits,
            )
            if not np.isnan(val):
                offset = np.where(cdiffmaglim[index] == val)[0][0]
                row["cmagpsf_ext"][offset] = val
                row["csigmapsf_ext"][offset] = 0.2

        out.append(row)

    return pd.Series(out)


def return_flatten_names(
    df: DataFrame, pref: str = "", flatten_schema: list = None
) -> list:
    """Retrieve full path from a nested schema

    Example, if one has a nested structure such as:
    root
     |-- timestamp: timestamp (nullable = true)
     |-- decoded: struct (nullable = true)
     |    |-- schemavsn: string (nullable = true)
     |    |-- publisher: string (nullable = true)
     |    |-- objectId: string (nullable = true)
     |    |-- candid: long (nullable = true)
     |    |-- candidate: struct (nullable = true)

    It will return a list like
        ["timestamp", "decoded" ,"decoded.schemavsn", "decoded.publisher", ...]

    Parameters
    ----------
    df : DataFrame
        Alert DataFrame
    pref : str, optional
        Internal variable to keep track of the structure, initially sets to "".
    flatten_schema: list, optional
        List containing the names of the flatten schema names.
        Initially sets to [].

    Returns
    -------
    flatten_frame: list
        List containing the names of the flatten schema names.

    Examples
    --------
    >>> df = spark.read.format("parquet").load("datatest")
    >>> flatten_schema = return_flatten_names(df)
    >>> assert("candidate.candid" in flatten_schema)
    """
    if (flatten_schema == []) or (flatten_schema is None):
        for colname in df.columns:
            flatten_schema.append(colname)  # noqa: PERF402

    # If the entry is not top level, it is then hidden inside a nested structure
    l_struct_names = [i.name for i in df.schema if isinstance(i.dataType, StructType)]

    for l_struct_name in l_struct_names:
        colnames = df.select("{}.*".format(l_struct_name)).columns
        for colname in colnames:
            if pref == "":
                flatten_schema.append(".".join([l_struct_name, colname]))
            else:
                flatten_schema.append(".".join([pref, l_struct_name, colname]))

        # Check if there are other levels nested
        flatten_schema = return_flatten_names(
            df.select("{}.*".format(l_struct_name)),
            pref=l_struct_name,
            flatten_schema=flatten_schema,
        )

    return flatten_schema


def apply_user_defined_filter(df: DataFrame, toapply: str, logger=None) -> DataFrame:
    """Apply a user filter to keep only wanted alerts.

    Parameters
    ----------
    df: DataFrame
        Spark DataFrame with alert data
    toapply: string
        Filter name to be applied. It should be in the form
        module.module.routine (see example below).

    Returns
    -------
    df: DataFrame
        Spark DataFrame with filtered alert data

    Examples
    --------
    >>> from pyspark.sql.functions import struct
    >>> colnames = ["cdsxmatch", "rb", "magdiff"]
    >>> df = spark.sparkContext.parallelize(zip(
    ...   ['RRLyr', 'Unknown', 'Star', 'SN1a'],
    ...   [0.01, 0.02, 0.6, 0.01],
    ...   [0.02, 0.05, 0.1, 0.01])).toDF(colnames)
    >>> df.show() # doctest: +NORMALIZE_WHITESPACE
    +---------+----+-------+
    |cdsxmatch|  rb|magdiff|
    +---------+----+-------+
    |    RRLyr|0.01|   0.02|
    |  Unknown|0.02|   0.05|
    |     Star| 0.6|    0.1|
    |     SN1a|0.01|   0.01|
    +---------+----+-------+
    <BLANKLINE>


    # Nest the DataFrame as for alerts
    >>> df = df.select(struct(df.columns).alias("candidate"))\
        .select(struct("candidate").alias("decoded"))

    # Apply quality cuts for example (level one)
    >>> toapply = 'fink_filters.filter_rrlyr.filter.rrlyr'
    >>> df = apply_user_defined_filter(df, toapply)
    >>> df.select("decoded.candidate.*").show() # doctest: +NORMALIZE_WHITESPACE
    +---------+----+-------+
    |cdsxmatch|  rb|magdiff|
    +---------+----+-------+
    |    RRLyr|0.01|   0.02|
    +---------+----+-------+
    <BLANKLINE>

    # Using a wrong filter name will lead to an error
    >>> df = apply_user_defined_filter(
    ...   df, "unknownfunc") # doctest: +SKIP
    """
    flatten_schema = return_flatten_names(df, pref="", flatten_schema=[])

    # Load the filter
    filter_name = toapply.split(".")[-1]
    module_name = toapply.split("." + filter_name)[0]
    module = importlib.import_module(module_name)
    filter_func = getattr(module, filter_name, None)

    # Note: to access input argument, we need f.func and not just f.
    # This is because f has a decorator on it.
    ninput = filter_func.func.__code__.co_argcount

    # Note: This works only with `struct` fields - not `array`
    argnames = filter_func.func.__code__.co_varnames[:ninput]
    colnames = []
    for argname in argnames:
        colname = [F.col(i) for i in flatten_schema if i.endswith("{}".format(argname))]
        if len(colname) == 0:
            raise AssertionError(
                """
                Column name {} is not a valid column of the DataFrame.
                """.format(argname)
            )
        colnames.append(colname[0])

    if logger is not None:
        logger.info(
            "new filter/topic registered: {} from {}".format(filter_name, module_name)
        )

    return (
        df.withColumn("toKeep", filter_func(*colnames))
        .filter("toKeep == true")
        .drop("toKeep")
    )


@pandas_udf(ArrayType(BooleanType()), PandasUDFType.SCALAR)
def apply_quality_flags_on_history(rb, nbad):
    """Apply quality flags for the history vector

    Parameters
    ----------
    rb: Series
        Pandas Series of list of float containing
        `rb` values from `prv_candidates`
    nbad: Series
        Pandas Series of list of int containing
        `nbad` values from `prv_candidates`

    Returns
    -------
    out: Series
        Pandas Series of list of boolean. True if good candidate
        False otherwise.

    Examples
    --------
    >>> df = spark.read.format("parquet").load("fink_utils/test_data/online")
    >>> df = df.withColumn(
    ...     'history_flag',
    ...     apply_quality_flags_on_history(
    ...         F.col("prv_candidates.rb"),
    ...         F.col("prv_candidates.nbad")))
    """
    rbf = rb.apply(lambda x: [i >= 0.55 for i in x])
    nbadf = nbad.apply(lambda x: [i == 0 for i in x])

    f = [np.array(i) * np.array(j) for i, j in zip(rbf.to_list(), nbadf.to_list())]
    return pd.Series(f)


def check_status_last_prv_candidates(df, status="valid"):
    """Check the `status` of the last alert in the history

    Parameters
    ----------
    df: Spark DataFrame
        Alert DataFrame (full schema)
    status: str
        Status can be:
            - valid: alert was processed by Fink
            - uppervalid: alert was not processed by Fink of quality cuts*
            - upper: upper limit from ZTF

    Returns
    -------
    out: Spark DataFrame
        Input DataFrame with an extra column (boolean) named `status`.

    Examples
    --------
    >>> df = spark.read.format("parquet").load("fink_utils/test_data/online")
    >>> print(df.count())
    11

    >>> df = check_status_last_prv_candidates(df, 'valid')
    >>> print(df.filter(F.col("valid")).count())
    11

    >>> df = check_status_last_prv_candidates(df, 'uppervalid')
    >>> print(df.filter(F.col("uppervalid")).count())
    0

    >>> df = check_status_last_prv_candidates(df, 'upper')
    >>> print(df.filter(F.col("upper")).count())
    0
    """
    # measurements are ordered from the most ancient to the newest
    # we want to check the last one
    rb = F.element_at("prv_candidates.rb", -1)
    nbad = F.element_at("prv_candidates.nbad", -1)
    magpsf = F.element_at("prv_candidates.magpsf", -1)
    if status == "valid":
        f1 = (rb >= 0.55) & (nbad == 0)
        cond = f1 & magpsf.isNotNull()
    elif status == "uppervalid":
        f1 = (rb >= 0.55) & (nbad == 0)
        cond = ~f1 & magpsf.isNotNull()
    elif status == "upper":
        cond = ~magpsf.isNotNull()

    return df.withColumn(status, cond)



# Copyright 2020-2024 AstroLab Software
# Author: Julien Peloton
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pandas as pd
import numpy as np
import pickle

from fink_utils.photometry.conversion import mag2fluxcal_snana


def format_data_as_snana(
    jd,
    measurement,
    error,
    fid,
    candid,
    mask,
    filter_conversion_dic=None,
    transform_to_flux=True,
):
    """Format data in SNANA units and format

    The resulting DataFrame is a concatenation of all alert data, with one
    measurement per row.

    |    a   |   b   |
    |--------|-------|
    | [1, 2] | [3, 4]|

    would become

    |    a   |   b   |
    |--------|-------|
    |    1   |   3   |
    |    2   |   4   |

    Parameters
    ----------
    jd: pd.Series
        Series containing Julian Dates (array of double). Each row contains
        all jd values for one alert (sorted).
    measurement: pd.Series
        Series containing data measurement (array of double). Each row contains
        all measurement values for one alert (sorted as jd).
        Can be either difference magnitude (a la ZTF), and which case you would
        set `transform_to_flux` to True to convert it into flux units,
        or it can be flux directly.
    error: pd.Series
        Series containing data error measurement (array of double). Each row
        contains all measurement values for one alert (sorted as jd).
        Can be either difference magnitude error (a la ZTF), and which case you
        would set `transform_to_flux` to True to convert it into flux units,
        or it can be flux error directly.
    fid: pd.Series
        Series containing filter band code (array of int/str). Each row contains
        all filter code values for one alert (sorted as jd).
    mask: pd.Series
        Series containing information on which alerts to keep (boolean).
    filter_conversion_dic: dict
        Mapping from telescope filter code (e.g. [1, 2] for ZTF) to
        SNANA filter code (['g', 'r']). Default is {1: 'g', 2: 'r', 3: 'i'}.
    transform_to_flux: boolean
        Set it to True if `measurement` is in Difference magnitude
        units (default for ZTF), in which case we will convert to apparent
        magnitude and then SNANA flux. Default is True.

    Returns
    -------
    pdf: pd.DataFrame
        DataFrame a la SNANA with SNID, MJD, FLUXCAL, FLUXCALERR, FLT.

    """
    if filter_conversion_dic is None:
        filter_conversion_dic = {1: "g", 2: "r", 3: "i"}
    # add an exploded column with SNID
    df_tmp = pd.DataFrame.from_dict({"jd": jd[mask], "SNID": candid[mask]})
    df_tmp = df_tmp.explode("jd")

    if transform_to_flux:
        # compute flux and flux error
        data = [
            mag2fluxcal_snana(*args)
            for args in zip(measurement[mask].explode(), error[mask].explode())
        ]
        flux, flux_error = np.transpose(data)
    else:
        flux = measurement[mask].explode()
        flux_error = error[mask].explode()

    # make a Pandas DataFrame with exploded series
    pdf = pd.DataFrame.from_dict({
        "SNID": df_tmp["SNID"],
        "MJD": df_tmp["jd"].astype("float"),
        "FLUXCAL": flux.astype("float"),
        "FLUXCALERR": flux_error.astype("float"),
        "FLT": fid[mask].explode().replace(filter_conversion_dic),
    })

    return pdf


def extract_history(history_list: list, field: str) -> list:
    """Extract the historical measurements contained in the alerts for the parameter `field`.

    Parameters
    ----------
    history_list: list of dict
        List of dictionary from alert[history].
    field: str
        The field name for which you want to extract the data. It must be
        a key of elements of history_list

    Returns
    -------
    measurement: list
        List of all the `field` measurements contained in the alerts.
    """
    if history_list is None:
        return []
    try:
        measurement = [obs[field] for obs in history_list]
    except KeyError:
        print("{} not in history data".format(field))
        measurement = []

    return measurement


def extract_field(alert: dict, field: str) -> np.array:
    """Concatenate current and historical observation data for a given field.

    Parameters
    ----------
    alert: dict
        Dictionnary containing alert data
    field: str
        Name of the field to extract.

    Returns
    -------
    data: np.array
        List containing previous measurements and current measurement at the
        end. If `field` is not in the category, data will be
        [alert['diaSource'][field]].
    """
    data = np.concatenate([
        [alert["candidate"][field]],
        extract_history(alert["prv_candidates"], field),
    ])
    return data


def load_scikit_model(fn: str = ""):
    """Load a RandomForestClassifier model from disk (pickled).

    Parameters
    ----------
    fn: str
        Filename. This file should be known from all machines!

    Return
    ----------
    clf: sklearn.ensemble.forest.RandomForestClassifier

    Examples
    --------
    >>> fn = 'fink_science/data/models/default-model_bazin.obj'
    >>> model = load_scikit_model(fn)
    >>> 'RandomForestClassifier' in str(type(model))
    True

    # binary classification
    >>> model.n_classes_
    2
    """
    return pickle.load(open(fn, "rb"))


def load_pcs(fn, npcs):
    """Load PC from disk into a Pandas DataFrame

    Parameters
    ----------
    fn: str
        Filename. This file should be known from all machines!
    npcs: int
        Number of principal components to load

    Return
    ----------
    pcs: pd.DataFrame
        All principal components to be considered.
        keys should be PCs names (1, 2, 3, ...),
        values their amplitude at each epoch in the grid.
        Order of PCs when calling pcs.keys() is important.
    """
    comp = pd.read_csv(fn)
    pcs = {}
    for i in range(npcs):
        pcs[i + 1] = comp.iloc[i].to_numpy()

    return pd.DataFrame(pcs)



from fink_utils.slack_bot.rich_text.rich_text import RichElement
from fink_utils.slack_bot.rich_text.rich_section import SectionElement
from enum import Enum
from typing import List, Union
from typing_extensions import Self


class RichListStyle(Enum):
    BULLET = "bullet"
    ORDERED = "ordered"


class RichList(RichElement):
    def __init__(
        self,
        style: RichListStyle = RichListStyle.BULLET,
        indent: int = None,
        offset: int = None,
        border: int = None,
    ) -> None:
        """
        A class representing a list

        Parameters
        ----------
        style : RichListStyle, optional
            style of the list, by default RichListStyle.BULLET
        indent : int, optional
            list indent in pixels, by default None
        offset : int, optional
            number of pixels to offset the list, by default None
        border : int, optional
            border thickness in pixels, by default None
        """
        super().__init__()
        self.rich_list = {
            "type": "rich_text_list",
            "style": style.value,
            "elements": [],
        }
        if indent is not None:
            self.rich_list["indent"] = indent

        if offset is not None:
            self.rich_list["offset"] = offset

        if border is not None:
            self.rich_list["border"] = border

    def get_element(self):
        return self.rich_list

    def add_elements(
        self, elements: Union[SectionElement, List[SectionElement]]
    ) -> Self:
        if type(elements) is list:
            self.rich_list["elements"] += [el.get_element() for el in elements]
        else:
            self.rich_list["elements"].append(elements.get_element())
        return self



from fink_utils.slack_bot.rich_text.rich_text import RichElement
from typing import List, Union
from typing_extensions import Self
from fink_utils.slack_bot.rich_text.rich_text_element import RichTextElement


class RichPreformatted(RichElement):
    def __init__(self, border: int = None) -> None:
        """
        A class representing a block surrounded by a border

        Parameters
        ----------
        border : _type_, optional
            border thickness in pixels, by default None
        """
        super().__init__()
        self.preform = {"type": "rich_text_preformatted", "elements": []}
        if border is not None:
            self.preform["border"] = border

    def get_element(self):
        return self.preform

    def add_elements(
        self, elements: Union[RichTextElement, List[RichTextElement]]
    ) -> Self:
        if type(elements) is list:
            self.preform["elements"] += [el.get_element() for el in elements]
        else:
            self.preform["elements"].append(elements.get_element())
        return self



# Copyright 2023-2024 AstroLab Software
# Author: Тимофей Пшеничный, Julien Peloton
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Part of these functionalities were taken from the Anomaly detection module"""

import os
import io
import time
import requests
import pandas as pd
import numpy as np
import gzip
from astropy.io import fits

import matplotlib.pyplot as plt


from fink_utils.data.utils import extract_field

COLORS_ZTF = {1: "#15284F", 2: "#F5622E"}


def status_check(res, header, sleep=8, timeout=25):
    """Checks whether the request was successful.

    In case of an error, sends information about the error to the @fink_test telegram channel

    Parameters
    ----------
    res : Response object

    Returns
    -------
        result : bool
            True : The request was successful
            False: The request was executed with an error
    """
    if res.status_code != 200:
        msg = """
        {}
        Content: {}
        """.format(header, res.content)
        url = "https://api.telegram.org/bot"
        url += os.environ["FINK_TG_TOKEN"]
        method = url + "/sendMessage"
        time.sleep(sleep)
        requests.post(
            method, data={"chat_id": "@fink_bot_error", "text": msg}, timeout=timeout
        )
        return False
    return True


def msg_handler_tg(tg_data, channel_id, init_msg, timeout=25):
    """Send `tg_data` to a telegram channel

    Notes
    -----
    The function sends notifications to the "channel_id" channel of Telegram.

    Parameters
    ----------
    tg_data: list
        List of tuples. Each item is a separate notification.
        Content of the tuple:
            text_data : str
                Notification text
            cutout : BytesIO stream
                cutout image in png format
            curve : BytesIO stream
                light curve picture
    channel_id: string
        Channel id in Telegram
    init_msg: str
        Initial message

    Returns
    -------
        None
    """
    url = "https://api.telegram.org/bot"
    url += os.environ["FINK_TG_TOKEN"]
    method = url + "/sendMediaGroup"

    if init_msg != "":
        res = requests.post(
            url + "/sendMessage",
            data={"chat_id": channel_id, "text": init_msg, "parse_mode": "markdown"},
            timeout=timeout,
        )
        status_check(res, header=channel_id)
    for text_data, cutout, curve in tg_data:
        files = {"first": curve}
        media = [
            {
                "type": "photo",
                "media": "attach://first",
                "caption": text_data,
                "parse_mode": "markdown",
            }
        ]
        if cutout is not None:
            files.update({"second": cutout})
            media.append({"type": "photo", "media": "attach://second"})
        res = requests.post(
            method,
            params={"chat_id": channel_id, "media": str(media).replace("'", '"')},
            files=files,
            timeout=timeout,
        )
        status_check(res, header=channel_id)
        time.sleep(10)


def msg_handler_tg_cutouts(tg_data, channel_id, init_msg, timeout=25, sleep_seconds=10):
    """Multi-cutout version of `msg_handler_tg`

    Notes
    -----
    The function sends notifications to the "channel_id" channel of Telegram.

    Parameters
    ----------
    tg_data: list
        List of tuples. Each item is a separate notification.
        Content of the tuple:
            text_data : str
                Notification text
            curve : BytesIO stream
                light curve picture
            cutouts : list of BytesIO stream
                List of cutout images in png format (1, 2, or 3 cutouts)
    channel_id: string
        Channel id in Telegram
    init_msg: str
        Initial message
    timeout: int
        Timeout when sending message. Default is 25 seconds.
    sleep_seconds: int
        How many seconds to sleep between two messages to avoid
        code 429 from the Telegram API. Default is 10 seconds.

    Returns
    -------
        None
    """
    url = "https://api.telegram.org/bot"
    url += os.environ["FINK_TG_TOKEN"]
    method = url + "/sendMediaGroup"

    if init_msg != "":
        res = requests.post(
            url + "/sendMessage",
            data={"chat_id": channel_id, "text": init_msg, "parse_mode": "markdown"},
            timeout=timeout,
        )
        status_check(res, header=channel_id)
    for text_data, curve, cutouts in tg_data:
        files = {"first": curve}
        media = [
            {
                "type": "photo",
                "media": "attach://first",
                "caption": text_data,
                "parse_mode": "markdown",
            }
        ]
        if isinstance(cutouts, list):
            names = ["second", "thrid", "fourth"]
            for cutout, name in zip(cutouts, names):
                files.update({name: cutout})
                media.append({"type": "photo", "media": "attach://{}".format(name)})
        res = requests.post(
            method,
            params={"chat_id": channel_id, "media": str(media).replace("'", '"')},
            files=files,
            timeout=timeout,
        )
        status_check(res, header=channel_id)
        time.sleep(sleep_seconds)


def get_cutout(cutout=None, ztf_id=None, kind="Difference", origin="alert"):
    """Loads cutout image from alert packet or via Fink API

    Parameters
    ----------
    cutout: bytes, optional
        Gzipped FITS file from alert packet. Only
        used for origin=alert.
    ztf_id : str, optional
        unique identifier for this object. Only
        used for origin=API.
    origin: str, optional
        Choose between `alert`[default], or API.

    Returns
    -------
    out : BytesIO stream
        cutout image in png format
    """
    if origin == "API":
        assert ztf_id is not None
        r = requests.post(
            "https://fink-portal.org/api/v1/cutouts",
            json={"objectId": ztf_id, "kind": kind, "output-format": "array"},
            timeout=25,
        )
        if not status_check(r, header=ztf_id):
            return io.BytesIO()
        data = np.log(np.array(r.json()["b:cutoutScience_stampData"], dtype=float))
        plt.axis("off")
        plt.imshow(data, cmap="PuBu_r")
        buf = io.BytesIO()
        plt.savefig(buf, format="png", bbox_inches="tight", pad_inches=0)
        buf.seek(0)
        plt.close()
    elif origin == "alert":
        assert cutout is not None

        # Unzip
        with gzip.open(io.BytesIO(cutout), "rb") as fits_file:
            with fits.open(
                io.BytesIO(fits_file.read()), ignore_missing_simple=True
            ) as hdul:
                img = hdul[0].data[::-1]

        data = np.log(img)
        plt.axis("off")
        plt.imshow(data, cmap="PuBu_r")
        buf = io.BytesIO()
        plt.savefig(buf, format="png", bbox_inches="tight", pad_inches=0)
        buf.seek(0)
        plt.close()
    else:
        buf = io.BytesIO()

    return buf


def get_curve(
    alert=None,
    jd=None,
    magpsf=None,
    sigmapsf=None,
    diffmaglim=None,
    fid=None,
    objectId=None,
    origin="API",
):
    """Generate PNG lightcurve

    Notes
    -----
    Based on `origin`, the other arguments are mandatory:
    - origin=API: objectId
    - origin=alert: alert
    - origin=fields: objectId, jd, magpsf, sigmapsf, diffmaglim, fid

    Parameters
    ----------
    alert: dict, optional
        alert packet containing `objectId`, `candidate`, `prv_candidates`
    jd: np.array, optional
        Vector of times.
    magpsf: np.array, optional
        Vector of difference magnitudes.
    sigmapsf: np.array, optional
        Vector of error on difference magnitudes.
    diffmaglim: np.array, optional
        Vector of upper limits on magnitude.
    fid: np.array, optional
        Vector of filter ID.
    objectId : str
        unique identifier for this object
    origin: str, optional
        Choose between `alert`, `API`[default], or `fields`.

    Returns
    -------
    out : BytesIO stream
        light curve picture
    """
    filter_dict = {1: "g band", 2: "r band"}
    if origin == "API":
        assert objectId is not None

        r = requests.post(
            "https://fink-portal.org/api/v1/objects",
            json={
                "objectId": objectId,
                "columns": "i:jd,i:fid,i:magpsf,i:sigmapsf,d:tag",
                "withupperlim": "True",
            },
        )
        if not status_check(r, header=objectId):
            return None

        # Format output in a DataFrame
        pdf = pd.read_json(io.BytesIO(r.content))

        plt.figure(figsize=(12, 4))

        for filt in pdf["i:fid"].unique():
            if filt == 3:
                continue
            maskFilt = pdf["i:fid"] == filt

            # The column `d:tag` is used to check data type
            maskValid = pdf["d:tag"] == "valid"
            plt.errorbar(
                pdf[maskValid & maskFilt]["i:jd"].apply(lambda x: x - 2400000.5),
                pdf[maskValid & maskFilt]["i:magpsf"],
                pdf[maskValid & maskFilt]["i:sigmapsf"],
                ls="",
                marker="o",
                color=COLORS_ZTF[filt],
                label=filter_dict[filt],
            )

            # see fink-utils#78 & fink-broker#872
            if "i:diffmaglim" in pdf.columns:
                maskUpper = pdf["d:tag"] == "upperlim"
                plt.plot(
                    pdf[maskUpper & maskFilt]["i:jd"].apply(lambda x: x - 2400000.5),
                    pdf[maskUpper & maskFilt]["i:diffmaglim"],
                    ls="",
                    marker="^",
                    color=COLORS_ZTF[filt],
                    markerfacecolor="none",
                )

            maskBadquality = pdf["d:tag"] == "badquality"
            plt.errorbar(
                pdf[maskBadquality & maskFilt]["i:jd"].apply(lambda x: x - 2400000.5),
                pdf[maskBadquality & maskFilt]["i:magpsf"],
                pdf[maskBadquality & maskFilt]["i:sigmapsf"],
                ls="",
                marker="v",
                color=COLORS_ZTF[filt],
            )

        plt.gca().invert_yaxis()
        plt.legend()
        plt.xlabel("Modified Julian Date")
        plt.ylabel("Difference magnitude")

        buf = io.BytesIO()
        plt.savefig(buf, format="png")
        buf.seek(0)
        plt.close()

    elif origin in ["alert", "fields"]:
        if origin == "alert":
            # extract current and historical data as one vector
            magpsf = extract_field(alert, "magpsf")
            sigmapsf = extract_field(alert, "sigmapsf")
            diffmaglim = extract_field(alert, "diffmaglim")
            fid = extract_field(alert, "fid")
            jd = extract_field(alert, "jd")
            objectId = alert["objectId"]

        # Rescale dates
        dates = np.array([i - jd[-1] for i in jd])

        # loop over filters
        plt.figure(num=1, figsize=(12, 4))

        # Loop over each filter
        for filt in COLORS_ZTF.keys():
            mask = np.where(fid == filt)[0]

            # Skip if no data
            if len(mask) == 0:
                continue

            # y data -- assume NaN (Spark style) and None (Pandas style) for missing values
            maskNotNone = np.array([
                ~np.isnan(i) and (i is not None) for i in magpsf[mask]
            ])
            plt.errorbar(
                dates[mask][maskNotNone],
                magpsf[mask][maskNotNone],
                yerr=sigmapsf[mask][maskNotNone],
                color=COLORS_ZTF[filt],
                marker="o",
                ls="",
                label=filter_dict[filt],
                mew=4,
            )
            # Upper limits
            plt.plot(
                dates[mask][~maskNotNone],
                diffmaglim[mask][~maskNotNone],
                color=COLORS_ZTF[filt],
                marker="v",
                ls="",
                mew=4,
                alpha=0.5,
            )
            plt.title(objectId)
        plt.legend()
        plt.gca().invert_yaxis()
        plt.xlabel("Days to candidates")
        plt.ylabel("Difference magnitude")

        buf = io.BytesIO()
        plt.savefig(buf, format="png")
        buf.seek(0)
        plt.close()

    return buf



# Copyright 2019 AstroLab Software
# Author: Julien Peloton
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Tuple
from pyspark import SparkContext
from pyspark.sql import SparkSession
from pyspark.sql import DataFrame
from pyspark.sql.column import Column, _to_java_column
from pyspark.sql.types import StructType

# import os
import json

from fink_utils.broker.avroUtils import readschemafromavrofile

# from fink_utils.test.tester import spark_unit_tests


def from_avro(dfcol: Column, jsonformatschema: str) -> Column:
    """Decode the Avro data contained in a DataFrame column into a struct.

    Notes
    -----
    Pyspark does not have all features contained in Spark core (Scala), hence
    we provide here a wrapper around the Scala function `from_avro`.
    You need to have the package org.apache.spark:spark-avro_2.11:2.x.y in the
    classpath to have access to it from the JVM.

    Parameters
    ----------
    dfcol: Column
        Streaming DataFrame Column with encoded Avro data (binary).
        Typically this is what comes from reading stream from Kafka.
    jsonformatschema: str
        Avro schema in JSON string format.

    Returns
    -------
    out: Column
        DataFrame Column with decoded Avro data.

    Examples
    --------
    >>> _, _, alert_schema_json = get_schemas_from_avro(ztf_avro_sample)
    >>> df_decoded = dfstream.select(
    ...   from_avro(dfstream["value"], alert_schema_json).alias("decoded"))
    >>> query = df_decoded.writeStream.queryName("qraw").format("memory")
    >>> t = query.outputMode("update").start()
    >>> t.stop()
    """
    sc = SparkContext._active_spark_context
    avro = sc._jvm.org.apache.spark.sql.avro
    f = getattr(getattr(avro, "package$"), "MODULE$").from_avro
    return Column(f(_to_java_column(dfcol), jsonformatschema))


def to_avro(dfcol: Column) -> Column:
    """Serialize the structured data of a DataFrame column into avro data (binary).

    Notes
    -----
    Since Pyspark does not have a function to convert a column to and from
    avro data, this is a wrapper around the scala function 'to_avro'.
    Just like the function above, to be able to use this you need to have
    the package org.apache.spark:spark-avro_2.11:2.x.y in the classpath.

    Parameters
    ----------
    dfcol: Column
        A DataFrame Column with Structured data

    Returns
    -------
    out: Column
        DataFrame Column encoded into avro data (binary).
        This is what is required to publish to Kafka Server for distribution.

    Examples
    --------
    >>> from pyspark.sql.functions import col, struct
    >>> avro_example_schema = '''
    ... {
    ...     "type" : "record",
    ...     "name" : "struct",
    ...     "fields" : [
    ...             {"name" : "col1", "type" : "long"},
    ...             {"name" : "col2", "type" : "string"}
    ...     ]
    ... }'''
    >>> df = spark.range(5)
    >>> df = df.select(struct("id",\
                 col("id").cast("string").alias("id2"))\
                 .alias("struct"))
    >>> avro_df = df.select(to_avro(col("struct")).alias("avro"))
    """
    sc = SparkContext._active_spark_context
    avro = sc._jvm.org.apache.spark.sql.avro
    f = getattr(getattr(avro, "package$"), "MODULE$").to_avro
    return Column(f(_to_java_column(dfcol)))


def write_to_csv(batchdf: DataFrame, batchid: int, fn: str = "web/data/simbadtype.csv"):
    """Write DataFrame data into a CSV file.

    The only supported Output Modes for File Sink is `Append`, but we need the
    complete table updated and dumped on disk here.
    Therefore this routine allows us to use CSV file sink with `Complete`
    output mode.
    TODO: that would be great to generalise this method!
    Get rid of these hardcoded paths!

    Parameters
    ----------
    batchdf: DataFrame
        Static Spark DataFrame with stream data. Expect 2 columns
        with variable names and their count.
    batchid: int
        ID of the batch.
    fn: str, optional
        Filename for storing the output.

    Examples
    --------
    >>> rdd = spark.sparkContext.parallelize(zip([1, 2, 3], [4, 5, 6]))
    >>> df = rdd.toDF(["type", "count"])
    >>> write_to_csv(df, 0, fn="test.csv")
    >>> os.remove("test.csv")
    """
    batchdf.toPandas().to_csv(fn, index=False)
    batchdf.unpersist()


def init_sparksession(
    name: str, shuffle_partitions: int = None, tz=None
) -> SparkSession:
    """Initialise SparkSession, the level of log for Spark and some configuration parameters

    Parameters
    ----------
    name: str
        Name for the Spark Application.
    shuffle_partitions: int, optional
        Number of partition to use when shuffling data.
        Typically better to keep the size of shuffles small.
        Default is None.
    tz: str, optional
        Timezone. Default is None.

    Returns
    -------
    spark: SparkSession
        Spark Session initialised.

    Examples
    --------
    >>> spark_tmp = init_sparksession("test")
    >>> conf = spark_tmp.sparkContext.getConf().getAll()
    """
    # Grab the running Spark Session,
    # otherwise create it.
    spark = SparkSession.builder.appName(name).getOrCreate()

    # keep the size of shuffles small
    if shuffle_partitions is not None:
        spark.conf.set("spark.sql.shuffle.partitions", shuffle_partitions)

    if tz is not None:
        spark.conf.set("spark.sql.session.timeZone", tz)

    # Set spark log level to WARN
    spark.sparkContext.setLogLevel("WARN")

    return spark


def get_spark_context() -> SparkContext:
    """Return the current SparkContext.

    Raise a RuntimeError if spark hasn't been initialized.

    Returns
    -------
    sparkContext : SparkContext instance
        The active sparkContext

    Examples
    --------
    >>> pysc = get_spark_context()
    >>> print(type(pysc))
    <class 'pyspark.context.SparkContext'>
    """
    if SparkContext._active_spark_context:
        return SparkContext._active_spark_context
    else:
        raise RuntimeError("SparkContext must be initialized")


def connect_to_kafka(
    servers: str,
    topic: str,
    startingoffsets: str = "latest",
    failondataloss: bool = False,
    kerberos: bool = False,
) -> DataFrame:
    """Initialise SparkSession, and set default Kafka parameters

    Parameters
    ----------
    servers: str
        kafka.bootstrap.servers as a comma-separated IP:PORT.
    topic: str
        Comma separated Kafka topic names.
    startingoffsets: str, optional
        From which offset you want to start pulling data. Options are:
        latest (only new data), earliest (connect from the oldest
        offset available), or a number (see Spark Kafka integration).
        Default is latest.
    failondataloss: bool, optional
        If True, Spark streaming job will fail if it is asking for data offsets
        that do not exist anymore in Kafka (because they have been deleted after
        exceeding a retention period for example). Default is False.
    kerberos: bool, optional
        If True, add options for a kerberized Kafka cluster. Default is False.

    Returns
    -------
    df: Streaming DataFrame
        Streaming DataFrame connected to Kafka stream

    Examples
    --------
    >>> dfstream_tmp = connect_to_kafka("localhost:29092", "ztf-stream-sim")
    >>> dfstream_tmp.isStreaming
    True
    """
    # Grab the running Spark Session
    spark = SparkSession.builder.getOrCreate()

    conf = spark.sparkContext.getConf().getAll()

    # Create a streaming DF from the incoming stream from Kafka
    df = spark.readStream.format("kafka").option("kafka.bootstrap.servers", servers)

    if kerberos:
        df = df.option(
            "kafka.sasl.kerberos.kinit.cmd",
            'kinit -t "%{sasl.kerberos.keytab}" -k %{sasl.kerberos.principal}',
        )
        df = df.option("kafka.sasl.kerberos.service.name", "kafka")

    # Naive check for secure connection - this can be improved...
    to_secure = sum(["-Djava.security.auth.login.config=" in i[1] for i in conf])  # noqa: C419
    if to_secure > 0:
        if kerberos:
            df = df.option("kafka.security.protocol", "SASL_PLAINTEXT").option(
                "kafka.sasl.mechanism", "GSSAPI"
            )
        else:
            df = df.option("kafka.sasl.mechanism", "PLAIN").option(
                "kafka.security.protocol", "SASL_SSL"
            )

    df = (
        df.option("subscribe", topic)
        .option("startingOffsets", startingoffsets)
        .option("failOnDataLoss", failondataloss)
        .load()
    )

    return df


def connect_to_raw_database(basepath: str, path: str, latestfirst: bool) -> DataFrame:
    """Initialise SparkSession, and connect to the raw database (Parquet)

    Parameters
    ----------
    basepath: str
        The base path that partition discovery should start with.
    path: str
        The path to the data (typically as basepath with a glob at the end).
    latestfirst: bool
        whether to process the latest new files first,
        useful when there is a large backlog of files

    Returns
    -------
    df: Streaming DataFrame
        Streaming DataFrame connected to the database

    Examples
    --------
    >>> dfstream_tmp = connect_to_raw_database(
    ...   "online/raw", "online/raw/*", True)
    >>> dfstream_tmp.isStreaming
    True
    """
    # Grab the running Spark Session
    spark = SparkSession.builder.getOrCreate()

    # Create a DF from the database
    userschema = spark.read.parquet(basepath).schema

    df = (
        spark.readStream.format("parquet")
        .schema(userschema)
        .option("basePath", basepath)
        .option("path", path)
        .option("latestFirst", latestfirst)
        .load()
    )

    return df


def load_parquet_files(path: str) -> DataFrame:
    """Initialise SparkSession, and load parquet files with Spark

    Unlike connect_to_raw_database, you get a standard DataFrame, and
    not a Streaming DataFrame.

    Parameters
    ----------
    path: str
        The path to the data

    Returns
    -------
    df: DataFrame
        Spark SQL DataFrame

    Examples
    --------
    >>> df = load_parquet_files(ztf_alert_sample)
    """
    # Grab the running Spark Session
    spark = SparkSession.builder.getOrCreate()

    # TODO: add mergeSchema option
    df = spark.read.format("parquet").option("mergeSchema", "true").load(path)

    return df


def get_schemas_from_avro(avro_path: str) -> Tuple[StructType, dict, str]:
    """Build schemas from an avro file (DataFrame & JSON compatibility)

    Parameters
    ----------
    avro_path: str
        Path to avro file from which schema will be extracted

    Returns
    -------
    df_schema: pyspark.sql.types.StructType
        Avro DataFrame schema
    alert_schema: dict
        Schema of the alert as a dictionary (DataFrame Style)
    alert_schema_json: str
        Schema of the alert as a string (JSON style)

    Examples
    --------
    >>> df_schema, alert_schema, alert_schema_json = get_schemas_from_avro(
    ...   ztf_avro_sample)
    >>> print(type(df_schema))
    <class 'pyspark.sql.types.StructType'>
    >>> print(type(alert_schema))
    <class 'dict'>
    >>> print(type(alert_schema_json))
    <class 'str'>
    """
    # Grab the running Spark Session
    spark = SparkSession.builder.getOrCreate()

    # Get Schema of alerts
    alert_schema = readschemafromavrofile(avro_path)
    df_schema = spark.read.format("avro").load("file://" + avro_path).schema
    alert_schema_json = json.dumps(alert_schema)

    return df_schema, alert_schema, alert_schema_json



# Copyright 2022-2024 AstroLab Software
# Author: Julien Peloton
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import requests
import subprocess
import json
import os

import pandas as pd
import numpy as np

from astropy.coordinates import SkyCoord
from astropy.time import Time
import astropy.constants as const
import astropy.units as u

from scipy import signal

from fink_utils.test.tester import regular_unit_tests


def query_miriade(
    ident, jd, observer="I41", rplane="1", tcoor=5, shift=15.0, timeout=30
):
    """Gets asteroid or comet ephemerides from IMCCE Miriade for a suite of JD for a single SSO

    Original function by M. Mahlke, adapted for Fink usage.

    Limitations:
        - Color ephemerides are returned only for asteroids
        - Temporary designations (C/... or YYYY...) do not have ephemerides available

    Parameters
    ----------
    ident: int, float, str
        asteroid or comet identifier
    jd: array
        dates to query
    observer: str
        IAU Obs code - default to ZTF: https://minorplanetcenter.net//iau/lists/ObsCodesF.html
    rplane: str
        Reference plane: equator ('1'), ecliptic ('2').
        If rplane = '2', then tcoor is automatically set to 1 (spherical)
    tcoor: int
        See https://ssp.imcce.fr/webservices/miriade/api/ephemcc/
        Default is 5 (dedicated to observation)
    shift: float
        Time shift to center exposure times, in second.
        Default is 15 seconds which is half of the exposure time for ZTF.
    timeout: int
        Timeout in seconds. Default is 30.

    Returns
    -------
    pd.DataFrame
        Input dataframe with ephemerides columns
        appended False if query failed somehow
    """
    # Miriade URL
    url = "https://ssp.imcce.fr/webservices/miriade/api/ephemcc.php"

    if rplane == "2":
        tcoor = "1"

    # Query parameters
    if ident.endswith("P") or ident.startswith("C/"):
        otype = "c"
    else:
        otype = "a"
    params = {
        "-name": f"{otype}:{ident}",
        "-mime": "json",
        "-rplane": rplane,
        "-tcoor": tcoor,
        "-output": "--jd,--colors(SDSS:r,SDSS:g)",
        "-observer": observer,
        "-tscale": "UTC",
    }

    # Pass sorted list of epochs to speed up query
    shift_hour = shift / 24.0 / 3600.0
    files = {
        "epochs": (
            "epochs",
            "\n".join(["{:.6f}".format(epoch + shift_hour) for epoch in jd]),
        )
    }

    # Execute query
    try:
        r = requests.post(url, params=params, files=files, timeout=timeout)
    except requests.exceptions.ReadTimeout:
        return pd.DataFrame()

    j = r.json()

    # Read JSON response
    try:
        ephem = pd.DataFrame.from_dict(j["data"])
    except KeyError:
        return pd.DataFrame()

    return ephem


def query_miriade_epehemcc(
    ident, jd, observer="I41", rplane="1", tcoor=5, shift=15.0, parameters=None
):
    """Gets asteroid or comet ephemerides from IMCCE Miriade for a suite of JD for a single SSO

    This uses local installation of ephemcc instead of the REST API.

    Limitations:
        - Color ephemerides are returned only for asteroids
        - Temporary designations (C/... or YYYY...) do not have ephemerides available

    Parameters
    ----------
    ident: int, float, str
        asteroid or comet identifier
    jd: array
        dates to query
    observer: str
        IAU Obs code - default to ZTF: https://minorplanetcenter.net//iau/lists/ObsCodesF.html
    rplane: str
        Reference plane: equator ('1'), ecliptic ('2').
        If rplane = '2', then tcoor is automatically set to 1 (spherical)
    tcoor: int
        See https://ssp.imcce.fr/webservices/miriade/api/ephemcc/
        Default is 5 (dedicated to observation)
    shift: float
        Time shift to center exposure times, in second.
        Default is 15 seconds which is half of the exposure time for ZTF.
    parameters: dict
        runner_path, userconf, iofile, outdir

    Returns
    -------
    pd.DataFrame
        Input dataframe with ephemerides columns
        appended False if query failed somehow

    """
    # write tmp files on disk
    date_path = "{}/dates_{}.txt".format(parameters["outdir"], ident)
    ephem_path = "{}/ephem_{}.json".format(parameters["outdir"], ident)

    pdf = pd.DataFrame(jd)

    shift_hour = shift / 24.0 / 3600.0
    pdf.apply(lambda epoch: Time(epoch + shift_hour, format="jd").iso).to_csv(
        date_path, index=False, header=False
    )

    # launch the processing
    cmd = [
        parameters["runner_path"],
        str(ident),
        str(rplane),
        str(tcoor),
        observer,
        "UTC",
        parameters["userconf"],
        parameters["iofile"],
        parameters["outdir"],
    ]

    # subprocess.run(cmd, capture_output=True)
    out = subprocess.run(cmd)

    if out.returncode != 0:
        print("Error code {}".format(out.returncode))

        # clean date file
        os.remove(date_path)

        return pd.DataFrame()

    # read the data from disk and return
    with open(ephem_path, "r") as f:
        data = json.load(f)
    ephem = pd.DataFrame(data["data"], columns=data["datacol"].keys())

    # clean tmp files
    os.remove(ephem_path)
    os.remove(date_path)

    return ephem


def get_miriade_data(
    pdf,
    sso_colname="i:ssnamenr",
    observer="I41",
    rplane="1",
    tcoor=5,
    withecl=True,
    method="rest",
    parameters=None,
    timeout=30,
):
    """Add ephemerides information from Miriade to a Pandas DataFrame with SSO lightcurve

    Parameters
    ----------
    pdf: pd.DataFrame
        Pandas DataFrame containing Fink alert data for a (or several) SSO
    sso_colname: str, optional
        Name of the column containing the SSO name to use when querying
        the miriade service. Beware that `ssnamenr` from ZTF is not very
        accurate, and it is better to use quaero before. Default is `i:ssnamenr`
    observer: str
        IAU Obs code - default to ZTF
        https://minorplanetcenter.net//iau/lists/ObsCodesF.html
    rplane: str
        Reference plane: equator ('1', default), ecliptic ('2').
        If rplane = '2', then tcoor is automatically set to 1 (spherical)
    tcoor: int
        See https://ssp.imcce.fr/webservices/miriade/api/ephemcc/
        Default is 5 (dedicated to observation)
    withecl: bool
        If True, query for also for ecliptic Longitude & Latitude (extra call to miriade).
        Default is True.
    method: str
        Use the REST API (`rest`), or a local installation of miriade (`ephemcc`)
    parameters: dict, optional
        If method == `ephemcc`, specify the mapping of extra parameters here. Default is {}.
    timeout: int, optional
        Timeout in seconds when using the REST API. Default is 30.

    Returns
    -------
    out: pd.DataFrame
        DataFrame of the same length, but with new columns from the ephemerides service.
    """
    if parameters is None:
        parameters = {}

    ssnamenrs = np.unique(pdf[sso_colname].values)

    infos = []
    for ssnamenr in ssnamenrs:
        mask = pdf[sso_colname] == ssnamenr
        pdf_sub = pdf[mask]

        if method == "rest":
            eph = query_miriade(
                str(ssnamenr),
                pdf_sub["i:jd"],
                observer=observer,
                rplane=rplane,
                tcoor=tcoor,
                timeout=timeout,
            )
        elif method == "ephemcc":
            eph = query_miriade_epehemcc(
                str(ssnamenr),
                pdf_sub["i:jd"],
                observer=observer,
                rplane=rplane,
                tcoor=tcoor,
                parameters=parameters,
            )
        else:
            raise AssertionError(
                "Method must be `rest` or `ephemcc`. {} not supported".format(method)
            )

        if not eph.empty:
            sc = SkyCoord(eph["RA"], eph["DEC"], unit=(u.deg, u.deg))

            eph = eph.drop(columns=["RA", "DEC"])
            eph["RA"] = sc.ra.value * 15
            eph["Dec"] = sc.dec.value

            if withecl:
                # Add Ecliptic coordinates
                if method == "rest":
                    eph_ec = query_miriade(
                        str(ssnamenr),
                        pdf_sub["i:jd"],
                        observer=observer,
                        rplane="2",
                        timeout=timeout,
                    )
                elif method == "ephemcc":
                    eph_ec = query_miriade_epehemcc(
                        str(ssnamenr),
                        pdf_sub["i:jd"],
                        observer=observer,
                        rplane="2",
                        parameters=parameters,
                    )
                else:
                    raise AssertionError(
                        "Method must be `rest` or `ephemcc`. {} not supported".format(
                            method
                        )
                    )

                sc = SkyCoord(
                    eph_ec["Longitude"], eph_ec["Latitude"], unit=(u.deg, u.deg)
                )
                eph["Longitude"] = sc.ra.value
                eph["Latitude"] = sc.dec.value

            # Merge fink & Eph
            info = pd.concat([eph.reset_index(), pdf_sub.reset_index()], axis=1)

            # index has been duplicated obviously
            info = info.loc[:, ~info.columns.duplicated()]

            # Compute magnitude reduced to unit distance
            info["i:magpsf_red"] = info["i:magpsf"] - 5 * np.log10(
                info["Dobs"] * info["Dhelio"]
            )
            infos.append(info)
        else:
            infos.append(pdf_sub)

    if len(infos) > 1:
        info_out = pd.concat(infos)
    else:
        info_out = infos[0]

    return info_out


def is_peak(x, y, xpeak, band=50):
    """Estimate if `xpeak` corresponds to a true extremum for a periodic signal `y`

    Assuming `y` a sparse signal along `x`, we would first estimate
    the period of the signal assuming a sine wave. We would then generate
    predictions, and locate the extrema of the sine.

    But this first step would generate false positives:
    1. As `y` is sparse, some extrema will not coincide with measurements
    2. As the signal is not a perfect sine, the fitted signal might shift
        from the real signal after several periods.

    This function is an extremely quick and dirty attempt to reduce false
    positives by looking at the data around a fitted peak, and
    estimating if the peak is real:
    1. take a band around the peak, and look if data is present
    2. if data is present, check the data is above the mean

    Parameters
    ----------
    xpeak: int
        Candidate peak position
    x: array
        Array of times
    y: array
        Array of elongation
    band: optional, int
        Bandwidth in units of x

    Returns
    -------
    out: bool
        True if `xpeak` corresponds to the location of a peak.
        False otherwise.
    """
    xband = np.where((x > xpeak - band) & (x < xpeak + band))[0]
    if (len(xband) >= 10) and (np.mean(y[xband]) > np.mean(y)):
        return True
    return False


def get_num_opposition(elong, width=4):
    """Estimate the number of opposition according to the solar elongation

    Under the hood, it assumes `elong` is peroidic, and uses a periodogram.

    Parameters
    ----------
    elong: array
        array of solar elongation corresponding to jd
    width: optional, int
        width of peaks in samples.

    Returns
    -------
    nopposition: int
        Number of oppositions estimate

    Examples
    --------
    >>> import io
    >>> import requests
    >>> import pandas as pd

    >>> r = requests.post(
    ...     'https://fink-portal.org/api/v1/sso',
    ...     json={
    ...         'n_or_d': '8467',
    ...         'withEphem': True,
    ...         'output-format': 'json'
    ...     }
    ... )
    >>> pdf = pd.read_json(io.BytesIO(r.content))

    # estimate number of oppositions
    >>> noppositions = get_num_opposition(
    ...     pdf['Elong.'].values,
    ...     width=4
    ... )
    >>> assert noppositions == 3, "Found {} oppositions for 8467 instead of 3!".format(noppositions)
    """
    peaks, _ = signal.find_peaks(elong, width=4)
    return len(peaks)


def compute_light_travel_correction(jd, d_obs):
    """Compute the time with light travel corrected

    Parameters
    ----------
    jd: np.array
        Array of times (JD), in day
    d_obs: np.array
        Array of distance to the observer, in AU
    """
    c_speed = const.c.to("au/day").value
    jd_lt = jd - d_obs / c_speed
    return jd_lt


def estimate_axes_ratio(residuals, R):
    """Estimate the axes ratio of a SSO from the residuals of the sHG1G2 model and its oblateness R.

    Parameters
    ----------
    residuals: np.array
        Residuals (observed - model) of the SSO with sHG1G2 model
    R: float
        Oblateness parameter of the sHG1G2 model

    Returns
    -------
    a_b, a_c: float
        a/b and a/c axes ratio
    """
    # Estimate the amplitude of the lightcurve from residuals
    # Taken at 2 sigma
    amplitude = np.std(residuals) * 2.0

    # Estimate the a/b ratio
    a_b = 10 ** (0.4 * (amplitude * 2))

    # Estimate the a/c ratio (and force c<b)
    a_c = (a_b + 1) / (2 * R)
    if a_c < a_b:
        a_c = a_b

    return a_b, a_c


if __name__ == "__main__":
    """Execute the unit test suite"""

    # Run the Spark test suite
    regular_unit_tests(globals())



from fink_utils.slack_bot.section import Section, TypeText
from fink_utils.slack_bot.image import Image
from typing import List, Union
from typing_extensions import Self
from fink_utils.slack_bot.rich_text.rich_text import RichText


class Message:
    def __init__(self):
        self.blocks = {"blocks": []}

    def add_elements(
        self,
        elements: Union[
            Section, Image, RichText, List[Union[Image, Section, RichText]]
        ],
    ) -> Self:
        """
        Add elements in the message

        Parameters
        ----------
        elements : Union[Section, Image, List[Union[Image, Section]]]
            elements to add in the message.
            Can be a single elements or a list of elements

        Returns
        -------
        Self
            return instance of the current object to chain add operation
        """
        if type(elements) is list:
            self.blocks["blocks"] += [el.get_element() for el in elements]
        else:
            self.blocks["blocks"].append(elements.get_element())
        return self

    def add_divider(self) -> Self:
        """
        Add a divider in the message
        """
        self.blocks["blocks"].append({"type": "divider"})
        return self

    def add_header(self, header_txt: str, allow_emoji: bool = False) -> Self:
        """
        Add a header in the message

        Parameters
        ----------
        type_txt : TypeText
            type of the header text
        header_txt : str
            header text
        allow_emoji : bool, optional
            if True allow emoji in the header, by default False
        """
        self.blocks["blocks"].append({
            "type": "header",
            "text": {
                "type": TypeText.PLAIN_TXT.value,
                "text": header_txt,
                "emoji": allow_emoji,
            },
        })
        return self


class Divider:
    def __init__(self) -> None:
        """
        Instanciate a divider
        """
        self.divider = {"type": "divider"}

    def get_element(self):
        return self.divider


class Header:
    def __init__(self, header_txt: str, allow_emoji: bool = False) -> None:
        self.header = {
            "type": "header",
            "text": {
                "type": TypeText.PLAIN_TXT.value,
                "text": header_txt,
                "emoji": allow_emoji,
            },
        }

    def get_element(self):
        return self.header



# Copyright 2019 AstroLab Software
# Author: Julien Peloton
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for manipulating Avro data and schemas. Some routines borrowed from lsst-dm/alert_stream and adapted."""

import io
import fastavro

from fink_utils.test.tester import regular_unit_tests

__all__ = ["writeavrodata", "readschemadata", "readschemafromavrofile"]


def writeavrodata(json_data: dict, json_schema: dict) -> io._io.BytesIO:
    """Encode json into Avro format given a schema.

    Parameters
    ----------
    json_data : `dict`
        The JSON data containing message content.
    json_schema : `dict`
        The writer Avro schema for encoding data.

    Returns
    -------
    `_io.BytesIO`
        Encoded data.

    Examples
    --------
    >>> with open(ztf_alert_sample, mode='rb') as file_data:
    ...   data = readschemadata(file_data)
    ...   # Read the schema
    ...   schema = data.writer_schema
    ...   for record in data:
    ...     bytes = writeavrodata(record, schema)
    >>> print(type(bytes))
    <class '_io.BytesIO'>
    """
    bytes_io = io.BytesIO()
    fastavro.schemaless_writer(bytes_io, json_schema, json_data)
    return bytes_io


def readschemadata(bytes_io: io._io.BytesIO) -> fastavro._read.reader:
    """Read data that already has an Avro schema.

    Parameters
    ----------
    bytes_io : `_io.BytesIO`
        Data to be decoded.

    Returns
    -------
    `fastavro._read.reader`
        Iterator over records (`dict`) in an avro file.

    Examples
    --------
    Open an avro file, and read the schema and the records
    >>> with open(ztf_alert_sample, mode='rb') as file_data:
    ...   data = readschemadata(file_data)
    ...   # Read the schema
    ...   schema = data.writer_schema
    ...   # data is an iterator
    ...   for record in data:
    ...     print(type(record))
    <class 'dict'>
    """
    bytes_io.seek(0)
    message = fastavro.reader(bytes_io)
    return message


def readschemafromavrofile(fn: str) -> dict:
    """Reach schema from a binary avro file.

    Parameters
    ----------
    fn: str
        Input Avro file with schema.

    Returns
    -------
    schema: dict
        Dictionary (JSON) describing the schema.

    Examples
    --------
    >>> schema = readschemafromavrofile(ztf_alert_sample)
    >>> print(schema['version'])
    3.3
    """
    with open(fn, mode="rb") as file_data:
        data = readschemadata(file_data)
        schema = data.writer_schema
    return schema


if __name__ == "__main__":
    """Execute the test suite"""
    # Add sample file to globals
    globs = globals()
    globs["ztf_alert_sample"] = "fink_utils/test_data/template_schema_ZTF_3p3.avro"

    # Run the regular test suite
    regular_unit_tests(globs)



from abc import ABC, abstractmethod
from typing import List, Union
from typing_extensions import Self


class RichElement(ABC):
    @abstractmethod
    def __init__(self) -> None:
        pass

    @abstractmethod
    def get_element(self):
        pass


class RichText:
    def __init__(self) -> None:
        """
        A class representing a rich text block
        """
        self.rich_text = {"type": "rich_text", "elements": []}

    def add_elements(self, elements: Union[RichElement, List[RichElement]]) -> Self:
        if type(elements) is list:
            self.rich_text["elements"] += [el.get_element() for el in elements]
        else:
            self.rich_text["elements"].append(elements.get_element())
        return self

    def get_element(self):
        return self.rich_text



# Copyright 2019 AstroLab Software
# Author: Julien Peloton
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import sys
import doctest
import numpy as np


def regular_unit_tests(global_args: dict = None, verbose: bool = False):
    """Base commands for the regular unit test suite

    Include this routine in the main of a module, and execute:
    python3 mymodule.py
    to run the tests.
    It should exit gracefully if no error (exit code 0),
    otherwise it will print on the screen the failure.

    Parameters
    ----------
    global_args: dict, optional
        Dictionary containing user-defined variables to
        be passed to the test suite. Default is None.
    verbose: bool
        If True, print useful debug messages.
        Default is False.

    Examples
    --------
    Set "toto" to "myvalue", such that it can be used during tests:
    >>> globs = globals()
    >>> globs["toto"] = "myvalue"
    >>> regular_unit_tests(global_args=globs)
    """
    if global_args is None:
        global_args = globals()

    # Numpy introduced non-backward compatible change from v1.14.
    if np.__version__ >= "1.14.0":
        np.set_printoptions(legacy="1.13")

    sys.exit(doctest.testmod(globs=global_args, verbose=verbose)[0])


def spark_unit_tests_broker(
    global_args: dict = None, verbose: bool = False, withstreaming: bool = False
):
    """
    Base commands for the spark unit test suite(fink-broker version)

    Include this routine in the main of a module, and execute:
    python3 mymodule.py
    to run the tests.
    It should exit gracefully if no error (exit code 0),
    otherwise it will print on the screen the failure.

    Parameters
    ----------
    global_args: dict, optional
        Dictionary containing user-defined variables to
        be passed to the test suite. Default is None.
    verbose: bool
        If True, print useful debug messages.
        Default is False.

    Examples
    --------
    Set "toto" to "myvalue", such that it can be used during tests:
    >>> globs = globals()
    >>> globs["toto"] = "myvalue"
    >>> regular_unit_tests(global_args=globs)
    """
    if global_args is None:
        global_args = globals()

    from pyspark.sql import SparkSession
    from pyspark import SparkConf

    conf = SparkConf()
    confdic = {
        "spark.jars.packages": os.environ["FINK_PACKAGES"],
        "spark.jars": os.environ["FINK_JARS"],
        "spark.python.daemon.module": "coverage_daemon",
    }
    conf.setMaster("local[2]")
    conf.setAppName("fink_test")
    for k, v in confdic.items():
        conf.set(key=k, value=v)
    spark = SparkSession.builder.appName("fink_test").config(conf=conf).getOrCreate()

    # Reduce the number of suffled partitions
    spark.conf.set("spark.sql.shuffle.partitions", 2)

    global_args["spark"] = spark

    if withstreaming:
        dfstream = (
            spark.readStream.format("kafka")
            .option("kafka.bootstrap.servers", os.environ["KAFKA_IPPORT_SIM"])
            .option("subscribe", os.environ["KAFKA_TOPIC"])
            .option("startingOffsets", "earliest")
            .load()
        )
        global_args["dfstream"] = dfstream

    # Numpy introduced non-backward compatible change from v1.14.
    if np.__version__ >= "1.14.0":
        np.set_printoptions(legacy="1.13")

    sys.exit(doctest.testmod(globs=global_args, verbose=verbose)[0])


def spark_unit_tests_science(global_args: dict = None, verbose: bool = False):
    """Base commands for the Spark unit test suite (fink-science version)

    Include this routine in the main of a module, and execute:
    python3 mymodule.py
    to run the tests.
    It should exit gracefully if no error (exit code 0),
    otherwise it will print on the screen the failure.

    Parameters
    ----------
    global_args: dict, optional
        Dictionary containing user-defined variables to
        be passed to the test suite. Default is None.
    verbose: bool
        If True, print useful debug messages.
        Default is False.
    """
    if global_args is None:
        global_args = globals()

    from pyspark.sql import SparkSession
    from pyspark import SparkConf

    spark = SparkSession.builder.getOrCreate()

    conf = SparkConf()
    confdic = {"spark.python.daemon.module": "coverage_daemon"}

    if spark.version.startswith("2"):
        confdic.update({
            "spark.jars.packages": "org.apache.spark:spark-avro_2.11:{}".format(
                spark.version
            )
        })
    elif spark.version.startswith("3"):
        confdic.update({
            "spark.jars.packages": "org.apache.spark:spark-avro_2.12:{}".format(
                spark.version
            )
        })
    conf.setMaster("local[2]")
    conf.setAppName("fink_science_test")
    for k, v in confdic.items():
        conf.set(key=k, value=v)
    spark = (
        SparkSession.builder.appName("fink_science_test")
        .config(conf=conf)
        .getOrCreate()
    )

    global_args["spark"] = spark

    # Numpy introduced non-backward compatible change from v1.14.
    if np.__version__ >= "1.14.0":
        np.set_printoptions(legacy="1.13")

    sys.exit(doctest.testmod(globs=global_args, verbose=verbose)[0])



# Copyright 2022 AstroLab Software
# Author: Julien Peloton
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json

from pyspark.sql import functions as F
from pyspark.sql import DataFrame


def load_hbase_catalog_as_dict(path_to_catalog: str) -> dict:
    """Load HBase table catalog from disk as dictionary

    Parameters
    ----------
    path_to_catalog: str
        Path to the json file containing the HBase table catalog

    Returns
    -------
    dcat: dict
        Dictionary containing the table catalog
    rowkey: str
        Name of the rowkey
    """
    with open(path_to_catalog) as f:
        catalog = json.load(f)

    dcat = json.loads(catalog)
    rowkey = [i["col"] for i in dcat["columns"].values() if i["cf"] == "rowkey"][0]

    return dcat, rowkey


def select_columns_in_catalog(catalog: dict, cols: list) -> (dict, str):
    """Select only `cols` in the catalog

    Parameters
    ----------
    catalog: dict
        Dictionary containing the HBase table catalog
    cols: list of str
        List of column names to keep

    Returns
    -------
    dcat_small: dict
        Same as input, but with fewer columns
    catalog_small: str
        json str view of the above
    """
    dcat_small = {}

    for k in catalog.keys():
        if k != "columns":
            dcat_small[k] = catalog[k]
        else:
            dcat_small[k] = {k_: v_ for k_, v_ in catalog[k].items() if k_ in cols}

    catalog_small = json.dumps(dcat_small)

    return dcat_small, catalog_small


def group_by_key(df: DataFrame, key: str, position: int, sep="_") -> DataFrame:
    """Group by the input `df` by split(`key`, '_')[`position`]

    Parameters
    ----------
    df: DataFrame
        Input dataframe
    key: str
        Row to groupby against. Better performance if rowkey.
    position: int
        Nth element to return from the list after the split of `key` is performed
        Zero-based index: First is 0. -1 means last.
    sep: str
        Separator to use for splitting `key`. Default is `_`.

    Returns
    -------
    df_grouped: DataFrame
        Result of the group by with two columns `id`, `count`.
        `id` = split(`key`, '_')[`position`]
    """
    if position >= 0:
        # The position is not zero based internally, but 1 based index.
        position += 1
    # Groupby key
    df_grouped = (
        df.select(F.element_at(F.split(df[key], "_"), position).alias("id"))
        .groupby("id")
        .count()
    )

    return df_grouped



from PIL import Image
import pandas as pd

import io
import requests
import numpy as np
import matplotlib.pyplot as plt

import fink_utils.slack_bot.bot as slack_bot

from fink_science.image_classification.utils import img_normalizer
from fink_utils.logging.logs import init_logging
from fink_utils.slack_bot.msg_builder import Message, Divider, Header
from fink_utils.slack_bot.section import Section, TypeText
import fink_utils.slack_bot.image as img

from fink_utils.slack_bot.rich_text.rich_text import RichText
from fink_utils.slack_bot.rich_text.rich_section import SectionElement
from fink_utils.slack_bot.rich_text.rich_preformatted import RichPreformatted
from fink_utils.slack_bot.rich_text.rich_quote import RichQuote
from fink_utils.slack_bot.rich_text.rich_list import RichList
import fink_utils.slack_bot.rich_text.rich_text_element as rel


def unzip_img(stamp: bytes, title_name: str) -> io.BytesIO:
    """
    Unzip and prepare an image to be send on slack

    Parameters
    ----------
    stamp : bytes
        raw alert cutout
    title_name : str
        image title

    Returns
    -------
    io.BytesIO
        image within a byte class
    """
    min_img = 0
    max_img = 255
    img = img_normalizer(stamp, min_img, max_img)
    img = Image.fromarray(img).convert("L")

    _ = plt.figure(figsize=(15, 6))
    plt.imshow(img, cmap="gray", vmin=min_img, vmax=max_img)
    plt.title(title_name)
    buf = io.BytesIO()
    plt.savefig(buf, format="png")
    buf.seek(0)
    return buf


def get_imgs():
    def request(kind):
        r = requests.post(
            "https://fink-portal.org/api/v1/cutouts",
            json={
                "objectId": "ZTF23abjzkmx",
                "kind": kind,
                "output-format": "array",
            },
        )
        return r.json()["b:cutout{}_stampData".format(kind)]

    return (
        np.array(request("Science")),
        np.array(request("Template")),
        np.array(request("Difference")),
    )


if __name__ == "__main__":
    raw_science, raw_template, raw_difference = get_imgs()

    prep_science = unzip_img(raw_science, "Science")
    prep_template = unzip_img(raw_template, "Template")
    prep_difference = unzip_img(raw_difference, "Difference")

    logger = init_logging("test_slack_bot")
    slack_client = slack_bot.init_slackbot(logger)

    results_post = slack_bot.post_files_on_slack(
        slack_client,
        [prep_science, prep_template, prep_difference],
        ["scienceImage", "templateImage", "differenceImage"],
        sleep_delay=1,
    )

    msg = Message()
    msg.add_header("Fink Slack Test")
    msg.add_divider()

    first_section = Section(
        TypeText.MARKDOWN, "this is a test message from the fink-utils CI"
    )

    first_section.add_textfield(TypeText.MARKDOWN, "first field :stars:", True)
    first_section.add_textfield(TypeText.MARKDOWN, "second field :ringed_planet:", True)

    science_section = Section(TypeText.MARKDOWN, "Science")
    science_section.add_slack_image(results_post[0], "Science")

    template_section = Section(TypeText.MARKDOWN, "Template")
    template_section.add_slack_image(results_post[1], "Template")

    difference_section = Section(TypeText.MARKDOWN, "Difference")
    difference_section.add_slack_image(results_post[2], "Difference")

    msg.add_elements(first_section)
    msg.add_elements(science_section)
    msg.add_elements(template_section)
    msg.add_elements(difference_section)

    fink_section = Section(TypeText.MARKDOWN, "fink-science-portal")
    fink_section.add_urlbutton(
        TypeText.PLAIN_TXT,
        "fink-science-portal",
        "View on Fink :fink:",
        "https://fink-portal.org/ZTF23abjzkmx",
        True,
    )
    msg.add_elements(fink_section)

    img_sci = img.Image(results_post[0], "scienceImage")
    img_temp = img.Image(results_post[1], "scienceTemplate")
    img_diff = img.Image(results_post[2], "scienceDifference")
    header_img = Header("Image Test")
    msg.add_elements([header_img, img_sci, Divider(), img_temp, Divider(), img_diff])

    slack_bot.post_msg_on_slack(
        slack_client,
        "#bot_fink_utils_test",
        [msg],
        logger=logger,
        verbose=True,
    )

    msg = (
        Message()
        .add_header("Type_text Test")
        .add_divider()
        .add_elements(
            RichText()
            .add_elements(
                SectionElement()
                .add_elements(rel.Text("- text element\n"))
                .add_elements(rel.Text("- next text\n"))
            )
            .add_elements(
                SectionElement().add_elements(rel.Link("https://fink-broker.org/"))
            )
            .add_elements(
                RichList()
                .add_elements(SectionElement().add_elements(rel.Text("first element")))
                .add_elements(SectionElement().add_elements(rel.Text("second element")))
                .add_elements(SectionElement().add_elements(rel.Text("third element")))
            )
            .add_elements(
                RichPreformatted()
                .add_elements(rel.Text("A preformatted block\n"))
                .add_elements(rel.Link("https://fink-broker.org/"))
            )
            .add_elements(RichQuote().add_elements(rel.Text("and finally a quote")))
        )
    )

    slack_bot.post_msg_on_slack(
        slack_client,
        "#bot_fink_utils_test",
        [msg],
        logger=logger,
        verbose=True,
    )



from enum import Enum
from typing_extensions import Self


class TypeText(Enum):
    PLAIN_TXT = "plain_text"
    MARKDOWN = "mrkdwn"


class Section:
    def __init__(
        self, type_text: TypeText, section_text: str, allow_emoji: bool = False
    ) -> None:
        """Instanciate a section.

        A section has always a text and cannot be empty.

        Parameters
        ----------
        type_text : TypeText
            type of the text, either markdown or plain_text
        section_text : str
            text of this section
        allow_emoji : bool, optional
            if True allow emoji in the text, by default False
        """
        assert len(section_text) > 0, "error, section_text cannot be empty"
        self.section = {"type": "section"}
        self.add_text(type_text, section_text, allow_emoji)

    def add_text(self, type_txt: TypeText, txt: str, allow_emoji: bool = False):
        """Override the text set by the constructor

        Parameters
        ----------
        type_txt : TypeText
            Type of the text
        txt : str
            text to put in the section
        allow_emoji : bool, optional
            if True allow the emoji in the text, by default False
        """
        self.section["text"] = {"type": type_txt.value, "text": txt}
        if type_txt == TypeText.PLAIN_TXT:
            self.section["text"]["emoji"] = allow_emoji

    def add_textfield(
        self, type_txt: TypeText, txt: str, allow_emoji: bool = False
    ) -> Self:
        """Add a text field in the section.

        The first call create the field.
        The next calls add more field into the textfield.

        Parameters
        ----------
        type_txt : TypeText
            type of the text in the field
        txt : str
            text of the field
        allow_emoji : bool, optional
            if True allow the emoji in the text, by default False

        Returns
        -------
        Self
            return instance of the current object to chain add operation
        """
        field_txt = {"type": type_txt.value, "text": txt}
        if type_txt == TypeText.PLAIN_TXT:
            field_txt = {"emoji": allow_emoji}

        if "fields" in self.section:
            self.section["fields"].append(field_txt)
        else:
            self.section["fields"] = [field_txt]
        return self

    def add_slack_image(self, image_slack_url: str, image_txt: str):
        """Add an image in the section with a text.

        WARNING: Call multiple times this method in the same section will override the previous call.
        Create multiple sections to add multiple images.

        Parameters
        ----------
        image_slack_url : str
            url of the slack image
        image_txt : str
            text to add with the image
        """
        self.section["accessory"] = {
            "type": "image",
            "slack_file": {"url": image_slack_url},
            "alt_text": image_txt,
        }

    def add_url_image(self, image_url: str, image_txt: str):
        """Add an image from the url in the section.

        Parameters
        ----------
        image_url : str
            url of the image
        image_txt : str
            text to add with the image
        """
        self.section["accessory"] = {
            "type": "image",
            "image_url": image_url,
            "alt_text": image_txt,
        }

    def add_urlbutton(
        self,
        type_txt: TypeText,
        button_value: str,
        button_txt: str,
        url: str,
        allow_emoji: bool = False,
    ):
        """Add a button with a url, click on it will redirect to the url.

        Parameters
        ----------
        type_txt : TypeText
            type of the text
        button_value : str
            value of the button
        button_txt : str
            text in the button
        url : str
            redirection url
        allow_emoji : bool, optional
            if True allow the emoji in the text, by default False
        """
        self.section["accessory"] = {
            "type": "button",
            "text": {"type": type_txt.value, "text": button_txt, "emoji": allow_emoji},
            "value": button_value,
            "url": url,
            "action_id": "button-action",
        }

    def get_element(self):
        return self.section
